from openpyxl import load_workbook
from openpyxl import Workbook
from openpyxl.utils import get_column_letter
import logging
import json
import os
import datetime as dt
from statistics import mean
import decimal
from pathlib import Path
from binubuo import binubuo
from binubuo.BinubuoTemplate import BinubuoTemplate

class binubuoExcel(binubuo):
    def __init__(self, binubuokey):
        super().__init__(binubuokey)
        self.binubuokey = binubuokey
        self.max_sample_size = 100
        self.connected = False
        self.show_messages = True

    def setup_logging(self):
        # Create the logger
        self.logger = logging.getLogger(__name__)
        # Logger settings and levels
        self.logger.setLevel(logging.DEBUG)
        # Logger formatter and handler
        file_handler = logging.FileHandler(filename='binubuo.log', mode='a', encoding='utf8')
        formatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')
        file_handler.setFormatter(formatter)
        # Add the handler to the logger
        self.logger.addHandler(file_handler)

    def excel_line_count(self):
        return len([row for row in self.ws if not all([cell.value is None for cell in row])])

    def can_float(self, x):
        try:
            float(x)
            return True
        except:
            return False

    def is_date_string(self, str_in):
        date_fmts = ('%Y-%m-%dT%H:%M:%S.%f', '%Y-%m-%d %H:%M:%S', '%d-%m-%Y %H:%M:%S', '%Y-%m-%d', '%d-%m-%Y', '%d/%m/%Y', '%Y','%b %d, %Y','%b %d, %Y','%B %d, %Y','%B %d %Y','%m/%d/%Y','%m/%d/%y','%b %Y','%B%Y','%b %d,%Y')
        parsed = []
        if str_in is not None:
            for fmt in date_fmts:
                try:
                    t = dt.datetime.strptime(str(str_in), fmt)
                    parsed.append((str_in, fmt, t)) 
                    break
                except ValueError as err:
                    pass
        return parsed

    def templateFromFile(self, file_name, output_type="json"):
        self.m("Create template from Excel file [" + file_name + "].", tolog=True)
        table_template = BinubuoTemplate()
        # Open file and go through it.
        try:
            self.wb = load_workbook(filename = file_name)
            self.logger.info("File %s opened successfully", file_name)
            self.ws = self.wb.active
            # Row count
            row_count = self.excel_line_count()
            self.logger.info("File %s has %s rows with data.", file_name, row_count)
            # TODO: Build header check function here. For now header is not supported.
            self.current_has_header = False
            # Define store
            excel_sample_store = {}
            col_count = 0
            self.logger.debug("Start loading the sample store.")
            count = 0
            # We are looping column wise here
            for col in self.ws.iter_cols(min_row=1):
                col_count += 1
                self.logger.debug("setting up data store with generated headers")
                head_name = "c" + str(col_count)
                excel_sample_store[head_name] = []
                for cell in col:
                    count += 1
                    if count == int(self.binubuo_config.get('EXCEL', 'max_row_sample', fallback=100)):
                        break
                    else:
                        if not cell.value:
                            excel_sample_store[head_name].append(None)
                        else:
                            excel_sample_store[head_name].append(str(cell.value))
            # Now we have the sample store. Enumerate each to get infer data.
            self.logger.debug("Enumerate sample store for infer JSON generation")
            for count, idx in enumerate(excel_sample_store):
                self.logger.debug("Sample data for col %s: %s", idx, excel_sample_store[idx])
                infer_meta = {}
                self.logger.debug("Start creating infer JSON for column %s", idx)
                infer_meta["column_name"] = idx
                infer_meta["table_name"] = Path(file_name).stem
                # Nulls before checking for datatype
                self.logger.debug("before null check")
                infer_meta["column_nulls"] = sum(x is None for x in excel_sample_store[idx])
                if infer_meta["column_nulls"] > 0:
                    infer_meta["column_nullable"] = "Y"
                else:
                    infer_meta["column_nullable"] = "N"
                # Datatype check
                nums_chk = [x for x in excel_sample_store[idx] if self.can_float(x)]
                self.logger.debug("Numbers checked: %s", nums_chk)
                self.logger.debug("before datatype check")
                if len(nums_chk) == len(excel_sample_store[idx]) or len(nums_chk) == (len(excel_sample_store[idx]) - infer_meta["column_nulls"]):
                    infer_meta["column_datatype"] = "number"
                    infer_meta["column_number_decimals"] = round(mean([decimal.Decimal(str(x)).as_tuple().exponent for x in nums_chk])*-1)
                    self.logger.debug("Column %s is datatype: %s", idx, "number")
                else:
                    # TODO: Here we need to check if it is a date, before we just set it to string
                    # Checking for date format is very intensive. So as soon as we have one fail (len(0)) returned from the checker, we give up.
                    is_date = None
                    date_format = None
                    for str_num, str_val in enumerate(excel_sample_store[idx]):
                        self.logger.debug("About to date check %s", str_val)
                        has_date_format = self.is_date_string(str_val)
                        if len(has_date_format) == 0:
                            is_date = False
                            break
                        else:
                            is_date = True
                            date_format = has_date_format[0][1]
                            self.logger.debug("Found format: %s", date_format)
                    if is_date:
                        infer_meta["column_datatype"] = "date"
                        self.logger.debug("Column %s is datatype: %s", idx, "date")
                    else:
                        infer_meta["column_datatype"] = "string"
                        self.logger.debug("Column %s is datatype: %s", idx, "string")
                self.logger.debug("before data metadata settings")
                infer_meta["table_level_rowcount"] = row_count
                infer_meta["table_level_avg_row_length"] = ""
                infer_meta["column_level_avg_col_length"] = round(sum( map(len, map(str, excel_sample_store[idx])) ) / len(excel_sample_store[idx]))
                infer_meta["column_data_length"] = "4000"
                infer_meta["column_number_precision"] = ""
                # Unique list
                u_set = set(excel_sample_store[idx])
                u_list = (list(u_set))
                # Bug with unique. For now ignore
                #if len(u_list) == len(excel_sample_store[idx]):
                #    infer_meta["column_is_unique"] = "yes"
                #    infer_meta["column_distinct_count"] = len(excel_sample_store[idx])
                #else:
                infer_meta["column_is_unique"] = "no"
                infer_meta["column_distinct_count"] = len(excel_sample_store[idx]) - len(u_list)
                self.logger.debug("before data low/high settings")
                if infer_meta["column_datatype"] == "number":
                    infer_meta["column_low_value"] = min([float(x) for x in nums_chk])
                    infer_meta["column_high_value"] = max([float(x) for x in nums_chk])
                elif infer_meta["column_datatype"] == "date":
                    self.logger.debug("Just before date min/max")
                    infer_meta["column_low_value"] = min([dt.datetime.strptime(x, date_format) for x in excel_sample_store[idx] if x is not None]).strftime('%Y-%m-%d %H:%M:%S')
                    infer_meta["column_high_value"] = max([dt.datetime.strptime(x, date_format) for x in excel_sample_store[idx] if x is not None]).strftime('%Y-%m-%d %H:%M:%S')
                elif infer_meta["column_datatype"] == "string":
                    infer_meta["column_low_value"] = min([x for x in excel_sample_store[idx] if x is not None])
                    infer_meta["column_high_value"] = max([x for x in excel_sample_store[idx] if x is not None])
                if infer_meta["column_datatype"] == "number":
                    infer_meta["sample_values"] = nums_chk
                else:
                    infer_meta["sample_values"] = excel_sample_store[idx]
                self.logger.debug("Infer: %s", json.dumps(infer_meta))
                table_template.init_column(column_name=idx)
                x_response = self.infer_generator(json.dumps(infer_meta))
                self.logger.info("Infer result: %s", x_response)
                table_template.replace_column_from_json(idx, json.dumps(x_response))
            self.m("Template for excel file [" + file_name +"] successfully created.", indent=True, tolog=False)
            if output_type.upper() == "JSON":
                table_template.validate_template()
                table_template.complete_template()
                return table_template.template_JSON
            else:
                return table_template
        except Exception as e:
            self.logger.error("Error %s happened during read.", e)

    def dataset_from_excel(self, file_name):
        dset_name = Path(file_name).stem.lower()
        dset_template = self.templateFromFile(file_name)
        self.create_dataset(dset_name, dset_template)

    def dataset_to_excel(self, dataset_name, file_name, sheet_name=None, dataset_type="custom", dataset_category=None):
        self.m("Create new excel file [" + file_name + "] of Dataset  [" + dataset_name + "].", tolog=True)
        if (dataset_type.upper() == "CUSTOM" and self.dataset_exists(dataset_name)) or (self.standard_dataset_exists(dataset_name, dataset_category)):
            # Dataset exists and we can continue
            if not os.path.exists(file_name):
                # File does not exist, we can proceed.
                # Get the data
                excel_data = self.dataset(dataset_name = dataset_name, dataset_type = dataset_type, dataset_category = dataset_category)
                # create the workbook
                new_wb = Workbook()
                new_ws = new_wb.active
                if sheet_name is None:
                    new_ws.title = dataset_name
                else:
                    new_ws.title = sheet_name
                # Append the rows from the dataset
                for rw in excel_data:
                    new_ws.append(rw)
                # Now save the new file
                new_wb.save(file_name)
                self.m("New excel file [" + file_name + "] created successfully.", tolog=True)
            else:
                self.logger.error("File %s already exist. Please choose another filename or remove existing file first.", file_name)
        else:
            self.logger.error("Dataset requested %s does not exist.", dataset_name)

    def copy_csv(self, file_name, new_name="newsheet"
            , drop_target_if_exist=False, data_rows="source"):
        self.m("Create copy [" + new_name + "] of Excel file  [" + file_name + "].", tolog=True)
        dset_name = Path(file_name).stem
        can_create = False
        self.logger.debug("Before checking existence")
        if not self.dataset_exists(dset_name):
            self.dataset_from_excel(file_name)
        if new_name.upper() != "NEWSHEET" or new_name.upper() != "SHEET":
            # We are writing to a new file. Check if it exists.
            self.logger.debug("Check if %s already exists", new_name)
            if os.path.exists(new_name):
                if drop_target_if_exist:
                    # remove file
                    self.logger.debug("%s already exists, but allow dropped set. Removing file.", new_name)
                    os.remove(new_name)
                    can_create = True
            else:
                self.logger.debug("%s does not exists.", new_name)
                can_create = True
        else:
            # Using keyword location
            can_create = True
            self.logger.debug("Keyword location chosen: %s", new_name)
        if can_create:
            # Set the rows
            if data_rows == "source":
                self.drows(self.excel_line_count(file_name))
            else:
                self.drows(int(data_rows))
            # Get the data
            excel_data = self.dataset(dset_name)
            if new_name.upper() == "NEWSHEET":
                # Create a new sheet in the existing file.
                wb = load_workbook(filename = file_name)
                self.logger.info("File %s opened successfully", file_name)
                new_ws = wb.create_sheet(title=dset_name)
                # Append the rows from the dataset
                for rw in excel_data:
                    new_ws.append(rw)
                wb.save(file_name)
            elif new_name.upper() == "SHEET":
                # Append to existing sheet in existing file.
                wb = load_workbook(filename = file_name)
                self.logger.info("File %s opened successfully", file_name)
                ws = wb.active
                for rw in excel_data:
                    ws.append(rw)
                wb.save(file_name)
            else:
                # Create a new file
                new_wb = Workbook()
                new_ws = new_wb.active
                new_ws.title = dset_name
                # Append the rows from the dataset
                for rw in excel_data:
                    new_ws.append(rw)
                # Now save the new file
                new_wb.save(new_name)
            self.m("Copy [" + new_name + "] created successfully.", tolog=True)




from genericpath import isfile
import pkg_resources
import os
import configparser
from pathlib import Path

from .binubuo import binubuo
from .BinubuoTemplate import BinubuoTemplate

config_exists = False
config_location = "In house"

# Check if we have configuration somewhere else than home directory
current_dir_obj = Path(os.getcwd())
current_dir_config = current_dir_obj / "binubuo.conf"
if current_dir_config.exists():
    # Check for config file in current directory
    config_location = current_dir_config
    config_exists = True
if "HOME" in os.environ:
    home_dir_obj = Path(os.environ["HOME"])
    home_dir_config = home_dir_obj / "binubuo.conf"
    if home_dir_config.exists():
    # Check for config file in users HOME
        config_location = home_dir_config
        config_exists = True
if "BINUBUO_HOME" in os.environ:
    # Look for config file here
    bhome_obj = Path(os.environ["BINUBUO_HOME"])
    bhome_config = bhome_obj / "binubuo.conf"
    if bhome_config.exists():
        config_location = bhome_config
        config_exists = True

if config_exists:
    binubuo_config = configparser.ConfigParser()
    binubuo_config.read(config_location)
else:
    binubuo_config = configparser.ConfigParser()
    binubuo_config['OUTPUTS'] = {'dependency_warnings': 'yes'}

# Get packages available for us. We need to check if the pre-requisites for the different database
# options are available when importing them.
pkgs_installed = {pkg.key for pkg in pkg_resources.working_set}

# Oracle requirements
oracle_required = {'cx-oracle'}
oracle_missing = oracle_required - pkgs_installed

if oracle_missing:
    if binubuo_config.get('OUTPUTS', 'dependency_warnings', fallback='yes') == 'yes':
        print("Oracle client not installed. Please install cx_Oracle for Oracle support.")
else:
    from .binubuoOracle import binubuoOracle

# Postgres requirements
postgres_required = {'psycopg2'}
postgres_missing = postgres_required - pkgs_installed

if postgres_missing:
    # Check for binary also
    postgres_bin_required = {'psycopg2-binary'}
    postgres_bin_missing = postgres_bin_required - pkgs_installed
    if postgres_bin_missing:
        if binubuo_config['OUTPUTS']['dependency_warnings'] == 'yes':
            print("Postgres client not installed. Please install Psycopg2 for Postgres support.")
    else:
        from .binubuoPostgres import binubuoPostgres
else:
    from .binubuoPostgres import binubuoPostgres

# yugabyte requirements
yugabyte_required = {'psycopg2'}
yugabyte_missing = yugabyte_required - pkgs_installed

if yugabyte_missing:
    # Check for binary also
    yugabyte_bin_required = {'psycopg2-binary'}
    yugabyte_bin_missing = yugabyte_bin_required - pkgs_installed
    if yugabyte_bin_missing:
        if binubuo_config['OUTPUTS']['dependency_warnings'] == 'yes':
            print("yugabyte client not installed. Please install Psycopg2 for YugabyteDB support.")
    else:
        from .binubuoYugabyte import binubuoYugabyte
else:
    from .binubuoYugabyte import binubuoYugabyte

# SQL Server requirements
sqlserver_required = {'pyodbc'}
sqlserver_missing = sqlserver_required - pkgs_installed

if sqlserver_missing:
    if binubuo_config['OUTPUTS']['dependency_warnings'] == 'yes':
        print("SQL Server client not installed. Please install pyodbc for SQL Server support.")
else:
    from .binubuoSQLServer import binubuoSQLServer

# Excel requirements
excel_required = {'openpyxl'}
excel_missing = excel_required - pkgs_installed

if excel_missing:
    if binubuo_config['OUTPUTS']['dependency_warnings'] == 'yes':
        print("Excel library not installed. Please install openpyxl for Excel support.")
else:
    from .binubuoExcel import binubuoExcel

# Always prepare CSV
from .binubuoCSV import binubuoCSV


from threading import main_thread
import pyodbc
import sys
import json
import ast
import time
import logging
from binubuo import binubuo
from binubuo.BinubuoTemplate import BinubuoTemplate

# Remove annoying incompatible warnings
import warnings
warnings.filterwarnings("ignore", category=DeprecationWarning) 

class binubuoSQLServer(binubuo):
    def __init__(self, binubuokey, dbserver, dbname, dbfullConnection=None):
        super().__init__(binubuokey)
        self.binubuokey = binubuokey
        self.dbserver = dbserver
        self.dbname = dbname
        self.dbfullConnection = dbfullConnection
        self.max_sample_size = 100
        self.connected = False
        self.binuObj = binubuo(binubuokey)
        self.show_messages = True

    def setup_logging(self):
        # Create the logger
        self.logger = logging.getLogger(__name__)
        # Logger settings and levels
        self.logger.setLevel(logging.DEBUG)
        # Logger formatter and handler
        file_handler = logging.FileHandler(filename='binubuo.log', mode='a')
        formatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')
        file_handler.setFormatter(formatter)
        # Add the handler to the logger
        self.logger.addHandler(file_handler)

    def connect(self):
        try:
            if self.dbfullConnection is None:
                #self.connection = pyodbc.connect(Trusted_Connection='yes', driver = '{ODBC Driver 17 for SQL Server}',server = self.dbserver , database = self.dbname)
                #self.connection = pyodbc.connect(Trusted_Connection='yes', driver = '{SQL Server Native Client 11.0}',server = self.dbserver , database = self.dbname)
                self.connection = pyodbc.connect(Trusted_Connection='yes', driver = '{SQL Server}',server = self.dbserver , database = self.dbname)
            else:
                self.connection = self.dbfullConnection
            self.connected = True
            #self.connection.autocommit = True
            working_cursor = self.connection.cursor()
            stmt = self.load_resource("sqlserver_date_format.sql")
            self.logger.info(stmt)
            working_cursor.execute(stmt)
        except:
            self.m("Error: Failed to connect to database. Please make sure you have a connection before using other methods.")
            self.connection = None

    def connectSecondary(self, dbserver, dbname, dbfullConnection=None):
        try:
            if dbfullConnection is None:
                self.secondary_connection = pyodbc.connect(Trusted_Connection='yes', driver = '{SQL Server Native Client 11.0}',server = dbserver , database = dbname)
            else:
                self.secondary_connection = dbfullConnection
            self.connected_secondary = True
            working_cursor = self.connection.cursor()
            stmt = self.load_resource("sqlserver_date_format.sql")
            self.logger.info(stmt)
            working_cursor.execute(stmt)
        except:
            self.m("Error: Failed to connect to secondary/alternate database. Please check your connection details and try again.")
            self.secondary_connection = None

    def calculateSampleSize(self, table_name):
        sub_sample_cursor = self.connection.cursor()
        sub_sample_size_stmt = self.load_resource("sqlserver_table_rows.sql")
        sub_sample_cursor.execute(sub_sample_size_stmt, (table_name.upper(),))
        sub_sample_size_cal = sub_sample_cursor.fetchone()
        # Calculate the real sample size in percent
        if (int(sub_sample_size_cal[0]) > 0 and int(sub_sample_size_cal[0]) < self.max_sample_size):
            sub_sample_size_cal_cust = 99.99
        elif (int(sub_sample_size_cal[0]) > self.max_sample_size):
            # Calculate rough percentage of max sample size
            sub_sample_size_cal_cust = round((self.max_sample_size/int(sub_sample_size_cal[0])*100), 8)
        else:
            # In case of zero rows, we read all of nothing :)
            sub_sample_size_cal_cust = 99.99
        self.logger.info("Sample size of %s calculated for %s rows.", str(sub_sample_size_cal_cust), str(sub_sample_size_cal[0]))
        return sub_sample_size_cal_cust

    def tableSize(self, table_name):
        table_size_cursor = self.connection.cursor()
        table_size_size_stmt = self.load_resource("sqlserver_table_rows.sql")
        table_size_cursor.execute(table_size_size_stmt, (table_name.upper(),))
        table_size_size_cal = table_size_cursor.fetchone()
        table_size_size_cal = table_size_size_cal[0]
        self.logger.info("Table size of %s is %s rows.", table_name, str(table_size_size_cal))
        return table_size_size_cal

    def fixDataTypeForJSON(self, data_in):
        if str(type(data_in)) == "<class 'datetime.datetime'>":
            data_out = data_in.strftime("%Y-%m-%dT%H:%M:%SZ")
        elif str(type(data_in)) == "<class 'bytes'>":
            # Let us restrict the size of the sample value, since we are just getting 
            # bytes as hex it could be huge.
            data_out = data_in.hex()[:30]
        elif str(type(data_in)) == "<class 'decimal.Decimal'>":
            data_out = float(data_in)
        elif str(type(data_in)) == "<class 'str'>":
            # Substring so we don't fill in sample values of huge documents.
            if len(data_in) > 250:
                data_out = data_in[:250]
            else:
                data_out = data_in
        else:
            data_out = data_in
        return data_out

    def templateFromTable(self, table_name, use_comments=True, use_infer=False, use_sample_data=False, output_type="json"):
        self.m("Create template from table [" + table_name + "]. Infer=" + str(use_infer) + " Sampledata=" + str(use_sample_data) + ".", tolog=True)
        table_template = BinubuoTemplate()
        # Let us validate the table name before we continue.
        table_exist = self.table_exist(table_name)
        # There are a number of different ways to branch out and do the template.
        # 1. Just use generic defaults for datatype.
        # 2. Use comments if available, and fall back to defaults if a column does not have a comment.
        # 3. Call the webservice with _only_ metadata to infer the generator
        # 4. Call the webservice with metadata _and_ sample data to infer the generator.
        if table_exist:
            self.templateTable = table_name.upper()
            if not use_infer:
                self.m("Using simple method to create columns.", indent=True, tolog=False)
                # No webservice call to generate template.
                working_cursor = self.connection.cursor()
                generic_column_meta_sql = self.load_resource("sqlserver_column_meta_simple.sql")
                working_cursor.execute(generic_column_meta_sql, (table_name.upper(),table_name.upper(),))
                for cname, cdtype, cdgenerator, ccomment, crows, cnumnulls, cnumdist, cnullable, cavglen in working_cursor:
                    table_template.init_column(column_name=cname, column_type="generated", column_datatype=cdtype)
                    if ccomment is not None and use_comments:
                        table_template.set_column_attribute(cname, "generator", ccomment)
                    else:
                        table_template.set_column_attribute(cname, "generator", cdgenerator)
                # Once we are done, we can validate the template and return it.
                self.m("Template for table [" + table_name +"] successfully created.", indent=True, tolog=False)
                if output_type.upper() == "JSON":
                    table_template.validate_template()
                    table_template.complete_template()
                    return table_template.template_JSON
                else:
                    return table_template
            else:
                # Placeholder to run the infer code.
                self.m("Using infer method to create columns.", indent=True, tolog=False)
                if use_sample_data:
                    sub_sample_size_cal_cust = self.calculateSampleSize(table_name)
                working_cursor = self.connection.cursor()
                if use_sample_data:
                    columns_meta_stmt = self.load_resource("sqlserver_column_meta_plus_sample.sql")
                else:
                    columns_meta_stmt = self.load_resource("sqlserver_column_meta_plus_standard.sql")
                self.logger.debug("Statement to get column data: %s", columns_meta_stmt)
                working_cursor.execute(columns_meta_stmt, (table_name.upper(),table_name.upper(),table_name.upper(),))
                # Stupid SQLServer does not allow for "per column" json.
                # So we are getting all and then looping over every array member which will be a column
                all_cols = ""
                for row in working_cursor:
                    all_cols = all_cols + row[0]
                self.logger.debug("Full text before json conversion: %s", all_cols)
                all_cols = json.loads(all_cols)
                self.logger.debug("Full json: %s", all_cols)
                table_size = self.tableSize(table_name)
                for idx, value in enumerate(all_cols):
                    self.logger.info("Column to infer: %s", value["column_name"])
                    sqlserver_schema = value["table_schema"]
                    # Get table size
                    main_meta = ""
                    # Because of sql servers wonderful obfuscated dictionary data we have to fetch some of the statistics data
                    # from dbcc calls. And even then we are not sure we can get any data. So enclose in try-except
                    # DBCC STATS_HEADER
                    try:
                        stats_cursor = self.connection.cursor()
                        column_stats_header_stmt = self.load_resource("sqlserver_column_stats_header.sql").format(schema_name = value["table_schema"], table_name = table_name, column_name = value["column_name"])
                        stats_header = stats_cursor.execute(column_stats_header_stmt).fetchall()
                        value["column_level_avg_col_length"] = int(stats_header[0][6])
                        self.logger.debug("Fetched stats header for col: %s, avg_col_len is: %s", value["column_name"], value["column_level_avg_col_length"])
                    except:
                        # Let us just make sure to close the ***** cursor and reopen
                        self.logger.debug("Error fetching stats header for col: %s.", value["column_name"])
                        stats_cursor.close()
                    try:
                        stats_cursor.close()
                    except:
                        pass
                    # DBCC HISTOGRAM
                    try:
                        stats_cursor = self.connection.cursor()
                        column_stats_histogram_stmt = self.load_resource("sqlserver_column_stats_histogram.sql").format(schema_name = value["table_schema"], table_name = table_name, column_name = value["column_name"])
                        stats_histogram_rows = stats_cursor.execute(column_stats_histogram_stmt).fetchall()
                        # Loop across all rows and collect and calculate some basic stats
                        lowest_val = self.fixDataTypeForJSON(stats_histogram_rows[0][0])
                        highest_val = self.fixDataTypeForJSON(stats_histogram_rows[0][0])
                        rows_at_lowest = stats_histogram_rows[0][2]
                        lowest_val_value = None
                        for histo_row in stats_histogram_rows:
                            if lowest_val is None and highest_val is not None:
                                lowest_val_value = self.fixDataTypeForJSON(histo_row[0])
                            highest_val = self.fixDataTypeForJSON(histo_row[0])
                        # Just check if both low and high is None, and if yes then all null.
                        # Else set the null percentage.
                        if lowest_val is None and highest_val is None:
                            # all nulls
                            value["column_nulls"] = 100
                        elif lowest_val is None and highest_val is not None:
                            value["column_nulls"] = round((rows_at_lowest/table_size)*100)
                            value["column_low_value"] = None
                            value["column_high_value"] = highest_val
                        else:
                            value["column_low_value"] = lowest_val_value
                            value["column_high_value"] = highest_val
                        self.logger.debug("Fetched stats header for col: %s, avg_col_len is: %s", value["column_name"], value["column_level_avg_col_length"])
                    except:
                        # Let us just make sure to close the ***** cursor and reopen
                        self.logger.debug("Error fetching stats histogram for col: %s.", value["column_name"])
                        stats_cursor.close()
                    try:
                        stats_cursor.close()
                    except:
                        pass
                    # If we are allowed sample data. Now is the time to fetch individual samples
                    if use_sample_data:
                        self.m("Sampling data on column [" + value["column_name"] +"].", indent=True, tolog=False)
                        sub_sample_cursor = self.connection.cursor()
                        # Merge the sample data.
                        # First parse the main column metadata
                        main_meta = value
                        if sub_sample_size_cal_cust < 5:
                            sub_sample_stmt = self.load_resource("sqlserver_column_sample_data_top.sql").format(r_col_name=value["column_name"], r_schema_name=value["table_schema"], r_table_name=table_name)
                        else:
                            sub_sample_stmt = self.load_resource("sqlserver_column_sample_data.sql").format(r_col_name=value["column_name"], r_schema_name=value["table_schema"], r_table_name=table_name, r_table_sample_size=sub_sample_size_cal_cust)
                        self.logger.debug("Statement to get column (%s) sample data: %s", value["column_name"], sub_sample_stmt)
                        # We need to make sure that we have a sample here in sqlserver, since a lot of the other
                        # statistics data we get in other databases are not readily available here.
                        if table_size > 0 and int(value["column_nulls"]) < 100:
                            self.logger.debug("Sampling to be done.")
                            got_sample_rows = False
                            try_sample_rows = 0
                            sample_arr = []
                            while not got_sample_rows:
                                sample_rows = sub_sample_cursor.execute(sub_sample_stmt).fetchall()
                                try_sample_rows = try_sample_rows + 1
                                self.logger.debug("Sample run: %s", try_sample_rows)
                                if len(sample_rows) > 0:
                                    self.logger.debug("We got %s rows.", len(sample_rows))
                                    got_sample_rows = True
                                else:
                                    if try_sample_rows > 10:
                                        self.logger.debug("Tried sampling 10 times without any rows. Breaking loop")
                                        value["column_nulls"] = 100
                                        value["column_low_value"] = 'Not extractable'
                                        value["column_high_value"] = 'Not extractable'
                                        sample_arr = []
                                        sample_rows = []
                                        break
                            do_text_validation = False
                            if len(sample_rows) > 0:
                                for sample_value in sample_rows:
                                    self.logger.debug("Type is %s", type(sample_value[0]))
                                    sample_arr.append(self.fixDataTypeForJSON(sample_value[0]))
                                    if str(type(sample_value[0])) == "<class 'str'>":
                                        do_text_validation = True
                            # Before we put the array with sample data, 
                            # we still need to check for combined size.
                            if do_text_validation:
                                comb_len = 0
                                avg_len = 0
                                for val in sample_arr:
                                    comb_len = comb_len + len(val)
                                avg_len = comb_len/len(sample_arr)
                                if comb_len > 20000:
                                    # Take 32000 divided by avg size elements only
                                    trim_elements = round(20000/avg_len)
                                    sample_arr = sample_arr[:trim_elements]
                        # Now add the sample data as a real array.
                        main_meta["sample_values"] = sample_arr
                        # Clear sample array for next check
                        sample_arr = []
                        sample_rows = []
                        if main_meta["column_is_reference"] == 1:
                            self.m("Column [" + value["column_name"] +"] is foreign key. Create key distribution histogram.", indent=True, tolog=False)
                            rel_sample_size_cal_cust = self.calculateSampleSize(main_meta["reference_table"])
                            ref_schema = self.table_sql_schema(main_meta["reference_table"])
                            rel_sample_cursor = self.connection.cursor()
                            self.logger.debug("Parameters for relation histogram. col_name=%s, schema_name=%s, tab_name=%s, ref_col_name=%s, ref_col_table=%s, ref_col_schema=%s, ref_sample=%s", main_meta["column_name"], sqlserver_schema, table_name, main_meta["reference_column"], main_meta["reference_table"], ref_schema, rel_sample_size_cal_cust)
                            if rel_sample_size_cal_cust > 5:
                                rel_sample_stmt = self.load_resource("sqlserver_foreign_key_histogram.sql").format(col_name = main_meta["column_name"], schema_name=sqlserver_schema, tab_name = table_name, ref_col_name = main_meta["reference_column"], ref_tab_name = main_meta["reference_table"], ref_sample = rel_sample_size_cal_cust, ref_schema_name = ref_schema)
                            else:
                                rel_sample_stmt = self.load_resource("sqlserver_foreign_key_histogram_top.sql").format(col_name = main_meta["column_name"], schema_name=sqlserver_schema, tab_name = table_name, ref_col_name = main_meta["reference_column"], ref_tab_name = main_meta["reference_table"], ref_schema_name = ref_schema)
                            self.logger.info("Column is a foreign key. Build relation histogram")
                            self.logger.debug("Relation histogram query: %s", rel_sample_stmt)
                            rel_sample_cursor.execute(rel_sample_stmt)
                            self.logger.debug("Rows affected: %s", rel_sample_cursor.rowcount)
                            all_ref_histo = ""
                            all_rows = rel_sample_cursor.fetchall()
                            for ref_histo_row in all_rows:
                                all_ref_histo = all_ref_histo + ref_histo_row[0]
                            self.logger.debug("Foreign key histogram: ")
                            rel_sample_data_clob_str = all_ref_histo
                            self.logger.debug("Histogram: %s", rel_sample_data_clob_str)
                            if len(rel_sample_data_clob_str) > 0:
                                rel_sample_data_clob_str = ast.literal_eval(rel_sample_data_clob_str)
                                main_meta["reference_histogram"] = rel_sample_data_clob_str
                            else:
                                self.logger.debug("Got zero rows from FK histogram sample query. Happens when the number of pages in the table is low.")
                    else:
                        main_meta = value
                    # We now have every column and its metadata. Send to API endpoint to do intelligent infer.
                    # We get back the entire json of inferred column
                    # Init column
                    table_template.init_column(column_name=value["column_name"])
                    # Try calling infer endpoint
                    self.logger.debug("Type of data to infer: %s", type(main_meta))
                    self.logger.info("Infer: %s", json.dumps(main_meta))
                    x_response = self.infer_generator(json.dumps(main_meta))
                    self.logger.info(x_response)
                    # The response is the column json as a string format. Call replace_column_from_json
                    table_template.replace_column_from_json(value["column_name"], json.dumps(x_response))
                self.m("Template for table [" + table_name +"] successfully created.", indent=True, tolog=False)
                # Once we are done, we can validate the template and return it.
                if output_type.upper() == "JSON":
                    table_template.validate_template()
                    table_template.complete_template()
                    return table_template.template_JSON
                else:
                    return table_template

    def columnGeneratorComments(self, table_name, comments, override=False):
        if self.connected:
            comment_count_sql = """select
                    count(exp.value)
                from 
                    INFORMATION_SCHEMA.COLUMNS ISC
                    left outer join sys.extended_properties exp on object_id(ISC.table_name) = exp.major_id and isc.ordinal_position = exp.minor_id
                WHERE
                    isc.table_name = ?
                and
                    exp.class_desc = 'OBJECT_OR_COLUMN'
                and
                    exp.name = 'MS_DESCRIPTION';"""
            working_cursor = self.connection.cursor()
            working_cursor.execute(comment_count_sql, (table_name,))
            cursor_result = working_cursor.fetchone()
            print(cursor_result)
            self.m("Comments: " + str(cursor_result[0]))
            if (cursor_result[0] > 0 or not override) or (cursor_result[0] == 0):
                # Let us go ahead and set comment generators as requested.
                if (comments.find('=') > 0):
                    # Named notation
                    for cn in comments.split(","):
                        colname = cn.split("=")[0].strip()
                        colcomment = cn.split("=")[1].strip()
                        col_add_stmt = """EXEC sp_addextendedproperty   
                            @name = N'MS_DESCRIPTION',   
                            @value = ?,  
                            @level0type = N'Schema', @level0name = 'dbo',  
                            @level1type = N'Table',  @level1name = ?,  
                            @level2type = N'Column', @level2name = ?"""
                        working_cursor.execute(col_add_stmt, (colcomment, table_name, colname))
                        self.connection.commit()
                else:
                    # Just split and follow column order. Ignore any column outside of split length
                    ordered_col_cursor = self.connection.cursor()
                    ordered_col_sql = "select ORDINAL_POSITION, COLUMN_NAME from INFORMATION_SCHEMA.COLUMNS where table_name = ? order by ORDINAL_POSITION asc;"
                    ordered_col_cursor.execute(ordered_col_sql, (table_name,))
                    comment_cursor = self.connection.cursor()
                    col_row = ordered_col_cursor.fetchone()
                    while col_row:
                        self.m("Working on col: " + col_row[1])
                        try:
                            colcomment = comments.split(",")[col_row[0] - 1]
                            col_add_stmt = """EXEC sp_addextendedproperty   
                                @name = N'MS_DESCRIPTION',   
                                @value = ?,  
                                @level0type = N'Schema', @level0name = 'dbo',  
                                @level1type = N'Table',  @level1name = ?,  
                                @level2type = N'Column', @level2name = ?"""
                            self.m("Comment command: " + col_add_stmt + "\nWith args: " + colcomment + " " + table_name + " " + col_row[1])
                            comment_cursor.execute(col_add_stmt, (colcomment, table_name, col_row[1]))
                        except:
                            # Aint no comment for this column. Just ignore
                            self.m("Soemthing went wrong here")
                        col_row = ordered_col_cursor.fetchone()
                    self.connection.commit()

    def quick_fetch_table(self
            , table_name, use_comments=True, use_infer=False
            , use_sample_data=False, use_tuple_return=False
            , output_csv=False, output_type="screen", output_name="same"
            , rows="source"):
        table_exist = self.table_exist(table_name)
        if table_exist:
            self.template = BinubuoTemplate()
            self.templateTable = table_name
        else:
            self.logger.error("Table name entered does not exist. templateFromTable can only be used on existing tables.")
            quit()
        # Table exists
        if table_exist:
            quick_fetch_columns_array = []
            if not use_infer:
                # No webservice call to generate template.
                working_cursor = self.connection.cursor()
                generic_column_meta_sql = self.load_resource("sqlserver_column_meta_simple.sql")
                working_cursor.execute(generic_column_meta_sql, (table_name.upper(),table_name.upper(),))
                for cname, cdtype, cdgenerator, ccomment, crows, cnumnulls, cnumdist, cnullable, cavglen in working_cursor:
                    self.template.init_column(column_name=cname, column_type="generated", column_datatype=cdtype)
                    if ccomment is not None and use_comments:
                        quick_fetch_columns_array.append(ccomment)
                    else:
                        quick_fetch_columns_array.append(cdgenerator)
                # Once we are done, we can call the quick fetch.
                quick_fetch_columns = ",".join(quick_fetch_columns_array)
                # Before fetching, set rows if needed.
                if rows == "source":
                    self.drows(self.tableSize(table_name))
                else:
                    try:
                        self.drows(int(rows))
                    except:
                        self.drows()
                if use_tuple_return or output_type.lower() == "table":
                    tuple_value = "tuple"
                else:
                    tuple_value = "list"
                if output_csv:
                    self.csv(1)
                if output_name == "same":
                    output_name_real = table_name
                else:
                    output_name_real = output_name
                if output_type.lower() == "screen" or output_type.lower() == "table":
                    resp_cols = self.quick_fetch(quick_fetch_columns, tuple_value)
                    self.logger.info("Length is: " + str(len(resp_cols)))
                    self.logger.debug(resp_cols)
                    if output_type.lower() == "screen":
                        return resp_cols
                    else:
                        # We need to insert.
                        quick_insert_stmt = self.build_insert_statement(table_name, output_name_real)
                        quick_insert_cursor = self.connection.cursor()
                        # For sqlserver, we need to disable identity insert restriction first
                        tab_schema = self.table_sql_schema(table_name)
                        quick_insert_cursor.execute('set identity_insert '+ tab_schema + '.' + output_name_real + ' on')
                        quick_insert_cursor.executemany(quick_insert_stmt, resp_cols)
                        self.connection.commit()
                        # Re-enable identity insert restriction
                        quick_insert_cursor.execute('set identity_insert '+ tab_schema + '.' + output_name_real + ' off')
                elif output_type.lower() == "file":
                    self.quick_fetch_to_file(quick_fetch_columns, output_name_real)
                # Reset required vals
                self.csv()
                self.drows()

    def dataset_from_table(self, table_name, use_comments=True, use_infer=False, use_sample_data=False):
        dset_template = self.templateFromTable(table_name, use_comments, use_infer, use_sample_data)
        # Create or replace the dataset
        self.create_dataset(table_name, dset_template)

    def build_insert_statement(self, source_table, target_table=None):
        insert_stmt_cursor = self.connection.cursor()
        build_insert_stmt = self.load_resource("sqlserver_build_insert.sql")
        insert_stmt_cursor.execute(build_insert_stmt, (source_table,))
        cursor_result = insert_stmt_cursor.fetchone()
        # For sql server we need the schema
        if target_table is not None:
            tab_schema = self.table_sql_schema(target_table)
            insert_stmt = "insert into " + tab_schema + "." + target_table + " " + cursor_result[0]
        else:
            tab_schema = self.table_sql_schema(source_table)
            insert_stmt = "insert into " + tab_schema + "." + source_table + " " + cursor_result[0]
        insert_stmt = insert_stmt.replace("d_d_s_r", "?")
        self.logger.debug("Insert stmt: " + insert_stmt)
        return insert_stmt

    def table_sql_schema(self, table_name):
        #
        get_schema_stmt_cursor = self.connection.cursor()
        get_schema_stmt = self.load_resource("sqlserver_get_table_schema.sql")
        get_schema_stmt_cursor.execute(get_schema_stmt, (table_name,))
        cursor_result = get_schema_stmt_cursor.fetchone()
        cursor_result = cursor_result[0]
        return cursor_result

    def table_has_identity(self, table_name):
        #
        get_schema_stmt_cursor = self.connection.cursor()
        get_schema_stmt = self.load_resource("sqlserver_table_has_identity.sql")
        get_schema_stmt_cursor.execute(get_schema_stmt, (table_name,))
        cursor_result = get_schema_stmt_cursor.fetchone()
        cursor_result = cursor_result[0]
        return cursor_result

    def table_exist(self, table_name, use_secondary=False):
        if use_secondary and self.connected_secondary:
            use_conn = self.secondary_connection
        else:
            use_conn = self.connection
        assert_table_sql = self.load_resource("sqlserver_assert_table.sql")
        # SOURCE work
        assert_table_cursor = use_conn.cursor()
        self.logger.debug("Before cursor check source")
        table_exist = False
        assert_table_cursor.execute(assert_table_sql, (table_name,))
        cursor_result = assert_table_cursor.fetchone()
        if cursor_result[0] == 1:
            table_exist = True
            self.logger.debug("Source exists")
        else:
            self.logger.info("Source table name entered does not exist. copy_table can only be used on existing tables.")
        return table_exist

    def ddl_from_table(self, table_name, use_secondary=False):
        if use_secondary and self.connected_secondary:
            use_conn = self.secondary_connection
        else:
            use_conn = self.connection
        ddl_cursor = use_conn.cursor()
        ddl_stmt = self.load_resource("sqlserver_table_ddl.sql")
        ddl_cursor.execute(ddl_stmt, (table_name,))
        ddl_out_str = ddl_cursor.fetchone()
        ddl_out_str = ddl_out_str[0]
        return ddl_out_str

    def copy_table(self, source_table, target_table=None, copy_method="quickfetch"
            , drop_target_if_exist=False, alternate_dataset_name=False
            , use_comments=True, use_infer=False, use_sample_data=False
            , data_rows="source", target_db=None):
        if target_db is None:
            target_work_cursor = self.connection.cursor()
        else:
            target_work_cursor = self.secondary_connection.cursor()
        source_table_exist = False
        target_table_exist = True
        self.logger.debug("Before cursor check source")
        working_cursor = self.connection.cursor()
        source_table_exist = self.table_exist(source_table)
        if source_table_exist:
            self.template = BinubuoTemplate()
            self.templateTable = source_table
        else:
            self.logger.error("Source table name entered does not exist. copy_table can only be used on existing tables.")
            quit()
        # If target_table is none, we will create a copy with same name plus '_copy' appended to it.
        if target_table is None:
            loc_target_table = source_table + "_copy"
        else:
            loc_target_table = target_table
        self.m("Create copy [" + loc_target_table + "] of table [" + source_table + "]. Infer=" + str(use_infer) + " Sampledata=" + str(use_sample_data) + ".", tolog=True)
        try:
            if target_db is None:
                target_table_exist = self.table_exist(loc_target_table)
            else:
                target_table_exist = self.table_exist(loc_target_table, True)
            if target_table_exist:
                self.logger.warning("Target table exist cannot create:")
                target_table_exist = True
                if drop_target_if_exist:
                    remove_target_stmt = "drop table " + loc_target_table.lower()
                    self.logger.info("Dropping target with stmt: " + remove_target_stmt)
                    target_work_cursor.execute(remove_target_stmt)
                    if target_db is None:
                        self.connection.commit()
                    else:
                        self.secondary_connection.commit()
                    target_table_exist = False
            else:
                self.logger.info("Target table does not exist.")
                target_table_exist = False
        except:
            target_table_exist = False
        self.logger.debug("Just before main continue check")
        if source_table_exist and not target_table_exist:
            # Source is there and target is not there. Good to go.
            # First we create an empty copy table
            tab_schema = self.table_sql_schema(source_table)
            if target_db is None:
                table_ddl = "select * into " + tab_schema + "." + loc_target_table.lower() + " from " + tab_schema + "."  + source_table.lower() + " where 1=2"
            else:
                table_ddl = self.ddl_from_table(source_table)
                # Replace with target
                table_ddl = table_ddl.replace('.' + source_table.lower() + ' ', '.' + loc_target_table.lower() + ' ')
            self.logger.debug("Create table stmt: " + table_ddl)
            target_work_cursor.execute(table_ddl)
            if target_db is None:
                self.connection.commit()
            else:
                self.secondary_connection.commit()
            # Table created. Build insert stmt
            insert_stmt = self.build_insert_statement(source_table, loc_target_table)
            # Set the rows
            if data_rows == "source":
                # Get the rowcount in the source table.
                self.drows(self.tableSize(source_table))
            else:
                try:
                    self.drows(int(data_rows))
                except:
                    self.drows(int(10))
            # Get new rows.
            if copy_method == "quickfetch":
                rows_bind = self.quick_fetch_table(table_name=source_table, use_tuple_return=True)
                self.logger.info("Rows returned for insert: " + str(len(rows_bind)))
                tab_schema = self.table_sql_schema(loc_target_table)
                target_work_cursor.execute('set identity_insert '+ tab_schema + '.' + loc_target_table + ' on')
                target_work_cursor.executemany(insert_stmt, rows_bind)
                target_work_cursor.execute('set identity_insert '+ tab_schema + '.' + loc_target_table + ' off')
                if target_db is None:
                    self.connection.commit()
                else:
                    self.secondary_connection.commit()
            elif copy_method == "dataset":
                if not alternate_dataset_name:
                    # Expect dataset name to be same as source table.
                    dset_name = source_table
                else:
                    dset_name = alternate_dataset_name
                # Create the dataset, if it does not exist
                if not self.dataset_exists(dset_name):
                    self.dataset_from_table(dset_name, use_comments, use_infer, use_sample_data)
                # Call the dataset for the rows.
                rows_bind = self.dataset(dataset_name=dset_name, response_type="tuple")
                if rows_bind is not None:
                    self.logger.info("Rows returned for insert: " + str(len(rows_bind)))
                    tab_schema = self.table_sql_schema(loc_target_table)
                    target_work_cursor.execute('set identity_insert '+ tab_schema + '.' + loc_target_table + ' on')
                    target_work_cursor.executemany(insert_stmt, rows_bind)
                    target_work_cursor.execute('set identity_insert '+ tab_schema + '.' + loc_target_table + ' off')
                if target_db is None:
                    self.connection.commit()
                else:
                    self.secondary_connection.commit()
                self.m("Synthetic table [" + loc_target_table +"] successfully created.", indent=True, tolog=False)

    def insert_from_dataset(self, dataset_name, table_name
        , data_rows="source", category="custom", **webargs):
        table_exist = self.table_exist(table_name)
        do_insert = False
        if table_exist:
            if data_rows == "source":
                # Get the rowcount in the source table.
                self.drows(self.tableSize(table_name))
            else:
                try:
                    self.drows(int(data_rows))
                except:
                    self.drows(int(10))
            # Create cursor and insert statement
            insert_cursor = self.connection.cursor()
            insert_stmt = self.build_insert_statement(table_name)
            if category == "custom":
                if self.dataset_exists(dataset_name):
                    # Custom dataset does not exists
                    rows_bind = self.dataset(dataset_name=dataset_name, response_type="tuple", **webargs)
                    do_insert = True
                elif self.standard_dataset_exists(dataset_name) is not None and self.binubuo_config['DATABASE']['use_standard_datasets_as_source'] == 'yes':
                    if self.binubuo_config.get('DATABASE', 'use_standard_category_if_duplicate', fallback=None) is None:
                        std_dataset_cat = self.standard_dataset_exists(dataset_name)
                    else:
                        std_dataset_cat = self.binubuo_config.get('DATABASE', 'use_standard_category_if_duplicate')
                    rows_bind = self.dataset(dataset_name=dataset_name, dataset_type="standard", dataset_category=std_dataset_cat, response_type="tuple", **webargs)
                    do_insert = True
                else:
                    self.logger.error("Dataset does not exist. Unable to insert.")
            else:
                if self.standard_dataset_exists(dataset_name, category).upper() == category.upper():
                    # We can do standard fetch
                    rows_bind = self.dataset(dataset_name=dataset_name, dataset_type="standard", dataset_category=category, response_type="tuple", **webargs)
                    do_insert = True
                else:
                    self.logger.error("Dataset does not exist. Unable to insert.")
            if do_insert and rows_bind is not None:
                self.logger.info("Rows returned for insert: " + str(len(rows_bind)))
                tab_schema = self.table_sql_schema(table_name)
                has_ident = self.table_has_identity(table_name)
                if has_ident == 1:
                    insert_cursor.execute('set identity_insert '+ tab_schema + '.' + table_name + ' on')
                insert_cursor.executemany(insert_stmt, rows_bind)
                if has_ident == 1:
                    insert_cursor.execute('set identity_insert '+ tab_schema + '.' + table_name + ' off')
                self.connection.commit()
        else:
            self.logger.error("Table does not exist. Unable to insert.")

    def table_sample(self, table_name):
        # Print title
        print("**************************************************************************")
        print("** Table data sample: " + table_name.upper())
        print("**************************************************************************")
        # Get header data
        headers = []
        headers_trimmed = []
        headers_deco = []
        header_stmt = self.load_resource("sqlserver_columns_ordered.sql")
        working_cursor = self.connection.cursor()
        working_cursor.execute(header_stmt, (table_name,))
        for cname in working_cursor:
            headers.append(cname[0])
            headers_trimmed.append(cname[0][0:14])
        row_format = "{:>15}" * (len(headers))
        for x in range (len(headers)):
            headers_deco.append(" ==============")
        print(row_format.format(*headers_trimmed))
        print(row_format.format(*headers_deco))
        tab_schema = self.table_sql_schema(table_name)
        sample_stmt = self.load_resource("sqlserver_sample_select.sql").format(r_limit = '3', r_rows = ','.join(headers), r_schema = tab_schema, r_table = table_name)
        working_cursor.execute(sample_stmt)
        for rows in working_cursor:
            row_data = []
            for col_data in rows:
                row_data.append(str(col_data)[0:14])
            print(row_format.format(*row_data))

    def compare_table_to_dataset(self, table_name, dataset_name):
        self.table_sample(table_name)
        self.dataset_sample(dataset_name)

    def build_schema_dataset(self, schema_name, use_comments=True, use_infer=False, use_sample_data=False, alternate_setname=None):
        self.m("Create API dataset for schema [" + schema_name + "]. Infer=" + str(use_infer) + " Sampledata=" + str(use_sample_data) + ".", tolog=True)
        # Prepare object for superset JSON registration
        superset_json = {}
        if not alternate_setname:
            superset_json["Setname"] = schema_name
        else:
            superset_json["Setname"] = alternate_setname
        superset_json["tables"] = []
        # Grab the tables in the correct order.
        working_cursor = self.connection.cursor()
        hierachy_stmt = self.load_resource("sqlserver_table_hierachy_schema.sql")
        all_rows = working_cursor.execute(hierachy_stmt).fetchall()
        for tabschema, tab, dependson, level in all_rows:
            # No matter if we have an existing dataset or not
            # We need to add to the superset
            table_def = {}
            table_def["table_name"] = tab
            table_def["table_level"] = level
            table_def["table_size"] = self.tableSize(tab)
            # Add dependents for size ratio calculations
            an_array = []
            depends_cursor = self.connection.cursor()
            depends_stmt = self.load_resource("sqlserver_foreign_key_depends.sql")
            depends_cursor.execute(depends_stmt, (tab,))
            for dep_tab in depends_cursor:
                an_object = {}
                an_object["local_col"] = dep_tab[2]
                an_object["remote_dataset"] = dep_tab[0]
                an_object["remote_col"] = dep_tab[1]
                an_array.append(an_object)    
            table_def["depends_on"] = an_array
            superset_json["tables"].append(table_def)
            if not self.dataset_exists(tab):
                self.logger.info("Building dataset for [%s] at level [%s].", tab, level)
                self.dataset_from_table(tab, use_comments, use_infer, use_sample_data)
            else:
                self.logger.info("Dataset for [%s] already exists", tab)
        # Now we have the whoel superset. Ship and create
        # TODO: Send to superset register service.
        self.logger.info("Superset data to store: %s", json.dumps(superset_json))
        superset_result = self.create_superset(json.dumps(superset_json))
        self.m("API dataset for schema [" + schema_name + "] successfully created.", tolog=True)

    def table_from_dataset(self, dataset_name, table_name=None, rows=0, category="custom", **webargs):
        self.m("Create table from dataset [" + dataset_name + "]. Number of rows=" + str(rows) + ".", tolog=True)
        self.logger.debug("Table from dataset using category: %s", category)
        # First we make sure that we have an up-to-date cache of datasets and metadata
        self.list_datasets(1)
        # Set the do_build
        do_build = False
        # Set the use standard
        use_standard = False
        # Now check if the dataset exists
        if category == "custom":
            if self.dataset_exists(dataset_name):
                do_build = True
            elif self.standard_dataset_exists(dataset_name) is not None and self.binubuo_config['DATABASE']['use_standard_datasets_as_source'] == 'yes':
                do_build = True
                use_standard = True
            else:
                self.m("Error: Dataset [" + dataset_name + "] does not exist.", tolog=True)
        else:
            if self.standard_dataset_exists(dataset_name, category).upper() == category.upper():
                do_build = True
                use_standard = True
            else:
                self.m("Error: Dataset [" + dataset_name + "] does not exist.", tolog=True)
        if do_build:
            if table_name is None:
                table_name_set = dataset_name
            else:
                table_name_set = table_name
            # Now check if the table exists
            if self.table_exist(table_name_set):
                self.logger.warning("Table [%s] does already exist. Either change name of requested table or drop existing.", table_name_set)
                # Here we should do "if allow drop, then drop"
            else:
                self.logger.info("About to create table [%s] from dataset definition", table_name_set)
                for idx, val in enumerate(self.dict_cache['datasets']):
                    for key, value in val.items():
                        if use_standard:
                            if key == "DATASET_WEBSERVICE_NAME" and value.lower() == dataset_name.lower():
                                if category == "custom":
                                    if self.dict_cache['datasets'][idx]["DATASET_TYPE_NAME"] == "Standard Dataset":
                                        dataset_cols = self.dict_cache['datasets'][idx]["DATASET_COLS"]
                                else:
                                    if self.dict_cache['datasets'][idx]["DATASET_TYPE_NAME"] == "Standard Dataset" and self.dict_cache['datasets'][idx]["DATASET_CATEGORY_NAME"].upper() == category.upper():
                                        dataset_cols = self.dict_cache['datasets'][idx]["DATASET_COLS"]
                        else:
                            if key == "DATASET_WEBSERVICE_NAME" and value.lower() == dataset_name.lower():
                                dataset_cols = self.dict_cache['datasets'][idx]["DATASET_COLS"]
                self.logger.info("Columns found for create table: %s", str(len(dataset_cols)))
                create_table_stmt = "create table dbo." + table_name_set + " ("
                # Oracle datatype translation dictionary
                translation_dict = {'string': 'nvarchar(4000)'
                    , 'time': 'datetime'
                    , 'text': 'nvarchar(max)'
                    , 'number': 'numeric(18,2)'}
                # loop over all columns and create the ddl for the table.
                cols_translated = []
                for idx, val in enumerate(dataset_cols):
                    trsl_datatype = dataset_cols[idx]["COL_DATATYPE"]
                    self.logger.debug("Add column [%s] of type [%s]", dataset_cols[idx]["COL_NAME"], trsl_datatype)
                    for word, replacement in translation_dict.items():
                        trsl_datatype = trsl_datatype.replace(word, replacement)
                    cols_translated.append(dataset_cols[idx]["COL_NAME"].lower() + " " + trsl_datatype)
                # Now all columns are done. Add to the statement
                all_cols = ','.join(cols_translated)
                create_table_stmt = create_table_stmt + all_cols + ")"
                self.logger.debug("Creating table: %s", create_table_stmt)
                working_cursor = self.connection.cursor()
                working_cursor.execute(create_table_stmt)
                self.connection.commit()
                self.m("Table " + table_name_set + " from dataset [" + dataset_name + "] created successfully.", tolog=True)
                try:
                    if rows > 0:
                        self.insert_from_dataset(dataset_name, table_name_set, rows, category, **webargs)
                except:
                    self.logger.debug("Either zero rows supplied or rows parameter not a number.")


import psycopg2
import sys
import json
import logging
from binubuo import binubuo
from binubuo.BinubuoTemplate import BinubuoTemplate

class binubuoPostgres(binubuo):
    def __init__(self, binubuokey, dbname=None, dbuser=None, dbpwd=None, dbfullConnection=None):
        super().__init__(binubuokey)
        self.binubuokey = binubuokey
        self.dbuser = dbuser
        self.dbpwd = dbpwd
        self.dbname = dbname
        self.dbfullConnection = dbfullConnection
        self.max_sample_size = 100
        self.connected = False
        self.show_messages = True

    def setup_logging(self):
        # Create the logger
        self.logger = logging.getLogger(__name__)
        # Logger settings and levels
        self.logger.setLevel(logging.DEBUG)
        # Logger formatter and handler
        file_handler = logging.FileHandler(filename='binubuo.log', mode='a')
        formatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')
        file_handler.setFormatter(formatter)
        # Add the handler to the logger
        self.logger.addHandler(file_handler)

    def connect(self):
        try:
            if self.dbfullConnection is None:
                if self.dbpwd is None:
                    self.connection = psycopg2.connect(dbname=self.dbname, user=self.dbuser)
                else:
                    self.connection = psycopg2.connect(dbname=self.dbname, user=self.dbuser, password=self.dbpwd)
            else:
                self.connection = self.dbfullConnection
            self.connected = True
        except:
            self.logger.error("Error: Failed to connect to database. Please make sure you have a connection before using other methods.")
            self.connection = None

    def connectSecondary(self, dbname, dbuser, dbpwd=None, dbfullConnection=None):
        try:
            if dbfullConnection is None:
                if dbpwd is None:
                    self.secondary_connection = psycopg2.connect(dbname=dbname, user=dbuser)
                else:
                    self.secondary_connection = psycopg2.connect(dbname=dbname, user=dbuser, password=dbpwd)
            else:
                self.secondary_connection = dbfullConnection
            self.connected_secondary = True
        except:
            self.logger.error("Error: Failed to connect to secondary/alternate database. Please check your connection details and try again.")
            self.secondary_connection = None

    def columnGeneratorComments(self, table_name, comments, override=False):
        self.m("Create comments on table [" + table_name + "]. Comments=" + str(comments) + ".", tolog=True)
        if self.connected:
            comment_count_sql = self.load_resource("postgres_count_comments.sql")
            working_cursor = self.connection.cursor()
            working_cursor.execute(comment_count_sql, (table_name,))
            cursor_result = working_cursor.fetchone()
            self.logger.debug("Comments: " + str(cursor_result[0]))
            if (cursor_result[0] > 0 or not override) or (cursor_result[0] == 0):
                # Let us go ahead and set comment generators as requested.
                if (comments.find('=') > 0):
                    # Named notation
                    for cn in comments.split(","):
                        colname = cn.split("=")[0].strip()
                        colcomment = cn.split("=")[1].strip()
                        col_add_stmt = "comment on column " + table_name + "." + colname + " is '" + colcomment + "'"
                        working_cursor.execute(col_add_stmt)
                else:
                    # Just split and follow column order. Ignore any column outside of split length
                    ordered_col_cursor = self.connection.cursor()
                    ordered_col_sql = self.load_resource("postgres_comments_ordered.sql")
                    ordered_col_cursor.execute(ordered_col_sql, (table_name,))
                    for cid, cname in ordered_col_cursor:
                        self.logger.debug("Working on col: " + cname)
                        try:
                            colcomment = comments.split(",")[cid - 1]
                            col_add_stmt = "comment on column " + table_name + "." + cname + " is '" + colcomment + "';"
                            self.logger.debug("Comment command: " + col_add_stmt)
                            working_cursor.execute(col_add_stmt)
                        except:
                            # Aint no comment for this column. Just ignore
                            pass
                self.connection.commit()
                self.m("Comments for table [" + table_name +"] successfully created.", indent=True, tolog=False)

    def calculateSampleSize(self, table_name):
        sub_sample_cursor = self.connection.cursor()
        sub_sample_size_stmt = self.load_resource("postgres_table_rows.sql")
        sub_sample_cursor.execute(sub_sample_size_stmt, (table_name,))
        sub_sample_size_cal = sub_sample_cursor.fetchone()
        # Calculate the real sample size in percent
        if (int(sub_sample_size_cal[0]) > 0 and int(sub_sample_size_cal[0]) < self.max_sample_size):
            sub_sample_size_cal_cust = 99.99
        elif (int(sub_sample_size_cal[0]) > self.max_sample_size):
            # Calculate rough percentage of max sample size
            sub_sample_size_cal_cust = round((self.max_sample_size/int(sub_sample_size_cal[0])*100), 8)
            # There seem to be issues where table sizes are in the goldielok zone, where the number of rows and percentage will have
            # A high probability of getting zero rows, because of the way postgres calculates the pages to sample.
            # So for now, we do a dirty fix and say that if pct is less than 10, set to 10.
            if sub_sample_size_cal_cust < 10:
                sub_sample_size_cal_cust = 10
        else:
            # In case of zero rows, we read all of nothing :)
            sub_sample_size_cal_cust = 99.99
        self.logger.info("Sample size of " + str(sub_sample_size_cal_cust) + " calculated for " + str(sub_sample_size_cal[0]) + " rows.")
        return sub_sample_size_cal_cust

    def tableIsAnalyzed(self, table_name):
        table_analyze_check_cursor = self.connection.cursor()
        table_analyze_check_stmt = self.load_resource("postgres_table_analyzed.sql")
        table_analyze_check_cursor.execute(table_analyze_check_stmt, (table_name,))
        table_analyze_check_val = table_analyze_check_cursor.fetchone()
        is_analyzed = table_analyze_check_val[0]
        if is_analyzed is None:
            self.logger.error("Table [%s] is not analyzed. Binubuo depends on statistics, so please analyze table for full functionality.", table_name)
            if self.binubuo_config['DATABASE']['force_analyze'] == 'yes':
                self.logger.info("Parameter force_analyze is yes. Analyze table [%s].", table_name)
                table_analyze_do_cursor = self.connection.cursor()
                table_analyze_do_cursor.execute("analyze " + table_name + ";")
                self.connection.commit()
                is_analyzed = "analyzed now"
        else:
            self.logger.info("Table [%s] was last anayzed %s.", table_name, str(is_analyzed))
        return is_analyzed

    def tableSize(self, table_name):
        table_size_cursor = self.connection.cursor()
        if self.tableIsAnalyzed(table_name) is not None:
            table_size_size_stmt = self.load_resource("postgres_table_rows.sql")
        else:
            table_size_size_stmt = self.load_resource("postgres_table_rows_not_analysed.sql")
        table_size_cursor.execute(table_size_size_stmt, (table_name,))
        table_size_size_cal = table_size_cursor.fetchone()
        table_size_size_cal = table_size_size_cal[0]
        self.logger.info("Table size of %s is %s rows.", table_name, str(table_size_size_cal))
        return table_size_size_cal

    def templateFromTable(self, table_name, use_comments=True, use_infer=False, use_sample_data=False, output_type="json"):
        self.m("Create template from table [" + table_name + "]. Infer=" + str(use_infer) + " Sampledata=" + str(use_sample_data) + ".", tolog=True)
        table_template = BinubuoTemplate()
        table_exist = self.table_exist(table_name)
        if not table_exist:
            self.logger.error("Table name entered does not exist. templateFromTable can only be used on existing tables.")
            quit()
        # Table exists
        if table_exist:
            table_template = BinubuoTemplate()
            if not use_infer:
                self.m("Using simple method to create columns.", indent=True, tolog=False)
                working_cursor = self.connection.cursor()
                generic_column_meta_sql = self.load_resource("postgres_column_meta_simple.sql")
                working_cursor.execute(generic_column_meta_sql, (table_name,))
                for cname, cdtype, cdgenerator, ccomment, crows, cnumnulls, cnumdist, cnullable, cavglen in working_cursor:
                    table_template.init_column(column_name=cname, column_type="generated", column_datatype=cdtype)
                    if ccomment is not None and use_comments:
                        table_template.set_column_attribute(cname, "generator", ccomment)
                    else:
                        table_template.set_column_attribute(cname, "generator", cdgenerator)
                # Once we are done, we can validate the template and return it.
                self.m("Template for table [" + table_name +"] successfully created.", indent=True, tolog=False)
                if output_type.upper() == "JSON":
                    table_template.validate_template()
                    table_template.complete_template()
                    return table_template.template_JSON
                else:
                    return table_template
            else:
                if self.tableIsAnalyzed(table_name) is None:
                    self.m("Table " + table_name + " is not analyzed. Please run [analyze " + table_name + ";]", indent=True, tolog=False)
                self.m("Using infer method to create columns.", indent=True, tolog=False)
                if use_sample_data:
                    sub_sample_size_cal_cust = self.calculateSampleSize(table_name)
                working_cursor = self.connection.cursor()
                if use_sample_data:
                    columns_meta_stmt = self.load_resource("postgres_column_meta_plus_sample.sql")
                else:
                    columns_meta_stmt = self.load_resource("postgres_column_meta_plus_standard.sql")
                working_cursor.execute(columns_meta_stmt, (table_name, table_name,))
                for col_name, col_infer_meta in working_cursor:
                    self.logger.info("Column to infer: " + col_name)
                    self.logger.debug("Infer: " + col_infer_meta["table_name"])
                    main_meta = col_infer_meta
                    if use_sample_data:
                        self.m("Sampling data on column [" + col_name +"].", indent=True, tolog=False)
                        sub_sample_cursor = self.connection.cursor()
                        # Merge the sample data.
                        sub_sample_stmt = self.load_resource("postgres_column_sample_data.sql").format(r_col_name=col_name, r_table_name=table_name, r_table_sample_size=sub_sample_size_cal_cust)
                        sub_sample_cursor.execute(sub_sample_stmt)
                        sample_count = 0
                        for sample_data_array in sub_sample_cursor:
                            # Now add the sample data as a real array.
                            if sample_data_array[0] is not None:
                                sample_count = sample_count + 1
                                main_meta["sample_values"] = sample_data_array[0]
                        self.logger.info("Fetched %s rows for sample data.", sample_count)
                        if main_meta["column_is_reference"] == 1:
                            self.m("Column [" + col_name +"] is foreign key. Create key distribution histogram.", indent=True, tolog=False)
                            rel_sample_size_cal_cust = self.calculateSampleSize(main_meta["reference_table"])
                            rel_sample_cursor = self.connection.cursor()
                            rel_sample_stmt = self.load_resource("postgres_foreign_key_histogram.sql").format(col_name = main_meta["column_name"], tab_name = table_name, ref_col_name = main_meta["reference_column"], ref_tab_name = main_meta["reference_table"], ref_sample = rel_sample_size_cal_cust)
                            # Fill in the histogram data
                            self.logger.info("Column is a foreign key. Build relation histogram")
                            self.logger.debug("Relation histogram query: " + rel_sample_stmt)
                            rel_sample_cursor.execute(rel_sample_stmt)
                            for ref_sample_data_array in rel_sample_cursor:
                                if ref_sample_data_array[0] is not None:
                                    self.logger.debug("Foreign key histogram: ")
                                    self.logger.debug(ref_sample_data_array[0])
                                    main_meta["reference_histogram"] = ref_sample_data_array[0]
                    # We get back the entire json of inferred column
                    # Init column and call infer endpoint
                    table_template.init_column(column_name=col_name)
                    self.logger.debug("Infer: " + json.dumps(main_meta))
                    x_response = self.infer_generator(json.dumps(main_meta))
                    self.logger.debug(x_response)
                    # The response is the column json as a string format. Call replace_column_from_json
                    table_template.replace_column_from_json(col_name, json.dumps(x_response))
                # Once we are done, we can validate the template and return it.
                self.m("Template for table [" + table_name +"] successfully created.", indent=True, tolog=False)
                if output_type.upper() == "JSON":
                    table_template.validate_template()
                    table_template.complete_template()
                    return table_template.template_JSON
                else:
                    return table_template


    def quick_fetch_table(self
            , table_name, use_comments=True, use_infer=False
            , use_sample_data=False, use_tuple_return=False
            , output_csv=False, output_type="screen", output_name="same"
            , rows="source"):
        table_exist = self.table_exist(table_name)
        if table_exist:
            self.template = BinubuoTemplate()
            self.templateTable = table_name
        else:
            self.logger.error("Table name entered does not exist. templateFromTable can only be used on existing tables.")
            quit()
        # Table exists
        if table_exist:
            quick_fetch_columns_array = []
            if not use_infer:
                # No webservice call to generate template.
                working_cursor = self.connection.cursor()
                generic_column_meta_sql = self.load_resource("postgres_column_meta_simple.sql")
                working_cursor.execute(generic_column_meta_sql, (table_name,))
                for cname, cdtype, cdgenerator, ccomment, crows, cnumnulls, cnumdist, cnullable, cavglen in working_cursor:
                    self.template.init_column(column_name=cname, column_type="generated", column_datatype=cdtype)
                    if ccomment is not None and use_comments:
                        quick_fetch_columns_array.append(ccomment)
                    else:
                        quick_fetch_columns_array.append(cdgenerator)
                # Once we are done, we can call the quick fetch.
                quick_fetch_columns = ",".join(quick_fetch_columns_array)
                # Before fetching, set rows if needed.
                if rows == "source":
                    # TODO: Set to current rowcount
                    self.drows(self.tableSize(table_name))
                else:
                    try:
                        self.drows(int(rows))
                    except:
                        self.drows()
                if use_tuple_return or output_type.lower() == "table":
                    tuple_value = "tuple"
                else:
                    tuple_value = "list"
                if output_csv:
                    self.csv(1)
                if output_name == "same":
                    output_name_real = table_name
                else:
                    output_name_real = output_name
                if output_type.lower() == "screen" or output_type.lower() == "table":
                    resp_cols = self.quick_fetch(quick_fetch_columns, tuple_value)
                    self.logger.info("Length is: " + str(len(resp_cols)))
                    self.logger.debug(resp_cols)
                    if output_type.lower() == "screen":
                        return resp_cols
                    else:
                        # We need to insert.
                        quick_insert_stmt = self.build_insert_statement(table_name, output_name_real)
                        quick_insert_cursor = self.connection.cursor()
                        quick_insert_cursor.executemany(quick_insert_stmt, resp_cols)
                        self.connection.commit()
                elif output_type.lower() == "file":
                    self.quick_fetch_to_file(quick_fetch_columns, output_name_real)
                # Reset required vals
                self.csv()
                self.drows()

    def dataset_from_table(self, table_name, use_comments=True, use_infer=False, use_sample_data=False):
        dset_template = self.templateFromTable(table_name, use_comments, use_infer, use_sample_data)
        # Create or replace the dataset
        self.create_dataset(table_name, dset_template)

    def build_insert_statement(self, source_table, target_table=None):
        insert_stmt_cursor = self.connection.cursor()
        build_insert_stmt = self.load_resource("postgres_build_insert.sql")
        insert_stmt_cursor.execute(build_insert_stmt, (source_table,))
        cursor_result = insert_stmt_cursor.fetchone()
        if target_table is not None:
            insert_stmt = "insert into " + target_table + " " + cursor_result[1]
        else:
            insert_stmt = "insert into " + source_table + " " + cursor_result[1]
        insert_stmt = insert_stmt.replace("d_d_s_r", "%s")
        self.logger.debug("Insert stmt: " + insert_stmt)
        return insert_stmt

    def table_exist(self, table_name, use_secondary=False):
        if use_secondary and self.connected_secondary:
            use_conn = self.secondary_connection
        else:
            use_conn = self.connection
        assert_table_sql = self.load_resource("postgres_assert_table.sql")
        assert_table_cursor = use_conn.cursor()
        self.logger.debug("Before cursor check source")
        table_exist = False
        assert_table_cursor.execute(assert_table_sql, (table_name,))
        cursor_result = assert_table_cursor.fetchone()
        if cursor_result[0]:
            table_exist = True
            self.logger.debug("Source exists")
        else:
            self.logger.info("Source table name entered does not exist. copy_table can only be used on existing tables.")
        return table_exist

    def ddl_from_table(self, table_name, use_secondary=False):
        if use_secondary and self.connected_secondary:
            use_conn = self.secondary_connection
        else:
            use_conn = self.connection
        ddl_cursor = use_conn.cursor()
        ddl_stmt = self.load_resource("postgres_table_ddl.sql")
        ddl_cursor.execute(ddl_stmt, (table_name,))
        ddl_out_str = ddl_cursor.fetchone()
        ddl_out_str = ddl_out_str[0]
        return ddl_out_str

    def copy_table(self, source_table, target_table=None, copy_method="quickfetch"
            , drop_target_if_exist=False, alternate_dataset_name=False
            , use_comments=True, use_infer=False, use_sample_data=False
            , data_rows="source", target_db=None):
        if target_db is None:
            target_work_cursor = self.connection.cursor()
        else:
            target_work_cursor = self.secondary_connection.cursor()
        source_table_exist = False
        target_table_exist = True
        self.logger.debug("Before cursor check source")
        working_cursor = self.connection.cursor()
        source_table_exist = self.table_exist(source_table)
        if source_table_exist:
            self.template = BinubuoTemplate()
            self.templateTable = source_table
        else:
            self.logger.error("Source table name entered does not exist. copy_table can only be used on existing tables.")
            quit()
        # If target_table is none, we will create a copy with same name plus '_copy' appended to it.
        if target_table is None:
            loc_target_table = source_table + "_copy"
        else:
            loc_target_table = target_table
        self.m("Create copy [" + loc_target_table + "] of table [" + source_table + "]. Infer=" + str(use_infer) + " Sampledata=" + str(use_sample_data) + ".", tolog=True)
        try:
            if target_db is None:
                target_table_exist = self.table_exist(loc_target_table)
            else:
                target_table_exist = self.table_exist(loc_target_table, True)
            if target_table_exist:
                self.logger.warning("Target table exist cannot create:")
                target_table_exist = True
                if drop_target_if_exist:
                    remove_target_stmt = "drop table " + loc_target_table.lower()
                    self.logger.info("Dropping target with stmt: " + remove_target_stmt)
                    target_work_cursor.execute(remove_target_stmt)
                    if target_db is None:
                        self.connection.commit()
                    else:
                        self.secondary_connection.commit()
                    target_table_exist = False
            else:
                self.logger.info("Target table does not exist.")
                target_table_exist = False
        except:
            target_table_exist = False
        self.logger.debug("Just before main continue check")
        if source_table_exist and not target_table_exist:
            # Source is there and target is not there. Good to go.
            # First we create an empty copy table
            if target_db is None:
                table_ddl = "create table " + loc_target_table.lower() + " as select * from " + source_table.lower() + " where 1=2"
            else:
                table_ddl = self.ddl_from_table(source_table)
                # Replace with target
                table_ddl = table_ddl.replace('CREATE TABLE ' + source_table + '', 'CREATE TABLE ' + loc_target_table.upper() + '')
            self.logger.debug("Create table stmt: " + table_ddl)
            target_work_cursor.execute(table_ddl)
            if target_db is None:
                self.connection.commit()
            else:
                self.secondary_connection.commit()
            # Table created. Build insert stmt
            insert_stmt = self.build_insert_statement(source_table, loc_target_table)
            # Set the rows
            if data_rows == "source":
                # Get the rowcount in the source table.
                self.drows(self.tableSize(source_table))
            else:
                try:
                    self.drows(int(data_rows))
                except:
                    self.drows(int(10))
            # Get new rows.
            if copy_method == "quickfetch":
                rows_bind = self.quick_fetch_table(table_name=source_table, use_tuple_return=True)
                self.logger.info("Rows returned for insert: " + str(len(rows_bind)))
                target_work_cursor.executemany(insert_stmt, rows_bind)
                if target_db is None:
                    self.connection.commit()
                else:
                    self.secondary_connection.commit()
            elif copy_method == "dataset":
                if not alternate_dataset_name:
                    # Expect dataset name to be same as source table.
                    dset_name = source_table
                else:
                    dset_name = alternate_dataset_name
                # Create the dataset, if it does not exist
                if not self.dataset_exists(dset_name):
                    self.dataset_from_table(dset_name, use_comments, use_infer, use_sample_data)
                # Call the dataset for the rows.
                rows_bind = self.dataset(dataset_name=dset_name, response_type="tuple")
                if rows_bind is not None:
                    self.logger.info("Rows returned for insert: " + str(len(rows_bind)))
                    target_work_cursor.executemany(insert_stmt, rows_bind)
                if target_db is None:
                    self.connection.commit()
                else:
                    self.secondary_connection.commit()
                self.m("Synthetic table [" + loc_target_table +"] successfully created.", indent=True, tolog=False)

    def insert_from_dataset(self, dataset_name, table_name
        , data_rows="source", category="custom", **webargs):
        table_exist = self.table_exist(table_name)
        do_insert = False
        if table_exist:
            if data_rows == "source":
                # Get the rowcount in the source table.
                self.drows(self.tableSize(table_name))
            else:
                try:
                    self.drows(int(data_rows))
                except:
                    self.drows(int(10))
            # Create cursor and insert statement
            insert_cursor = self.connection.cursor()
            insert_stmt = self.build_insert_statement(table_name)
            if category == "custom":
                if self.dataset_exists(dataset_name):
                    # Custom dataset does not exists
                    rows_bind = self.dataset(dataset_name=dataset_name, response_type="tuple", **webargs)
                    do_insert = True
                elif self.standard_dataset_exists(dataset_name) is not None and self.binubuo_config['DATABASE']['use_standard_datasets_as_source'] == 'yes':
                    if self.binubuo_config.get('DATABASE', 'use_standard_category_if_duplicate', fallback=None) is None:
                        std_dataset_cat = self.standard_dataset_exists(dataset_name)
                    else:
                        std_dataset_cat = self.binubuo_config.get('DATABASE', 'use_standard_category_if_duplicate')
                    rows_bind = self.dataset(dataset_name=dataset_name, dataset_type="standard", dataset_category=std_dataset_cat, response_type="tuple", **webargs)
                    do_insert = True
                else:
                    self.logger.error("Dataset does not exist. Unable to insert.")
            else:
                if self.standard_dataset_exists(dataset_name, category).upper() == category.upper():
                    # We can do standard fetch
                    rows_bind = self.dataset(dataset_name=dataset_name, dataset_type="standard", dataset_category=category, response_type="tuple", **webargs)
                    do_insert = True
                else:
                    self.logger.error("Dataset does not exist. Unable to insert.")
            if do_insert and rows_bind is not None:
                self.logger.info("Rows returned for insert: " + str(len(rows_bind)))
                insert_cursor.executemany(insert_stmt, rows_bind)
                self.connection.commit()
        else:
            self.logger.error("Table does not exist. Unable to insert.")

    def table_sample(self, table_name):
        # Print title
        print("**************************************************************************")
        print("** Table data sample: " + table_name.upper())
        print("**************************************************************************")
        # Get header data
        headers = []
        headers_trimmed = []
        headers_deco = []
        header_stmt = self.load_resource("postgres_columns_ordered.sql")
        working_cursor = self.connection.cursor()
        working_cursor.execute(header_stmt, (table_name,))
        for cname in working_cursor:
            headers.append(cname[0])
            headers_trimmed.append(cname[0][0:14])
        row_format = "{:>15}" * (len(headers))
        for x in range (len(headers)):
            headers_deco.append(" ==============")
        print(row_format.format(*headers_trimmed))
        print(row_format.format(*headers_deco))
        sample_stmt = self.load_resource("postgres_sample_select.sql").format(r_rows = ','.join(headers), r_table = table_name, r_limit = '3')
        working_cursor.execute(sample_stmt)
        for rows in working_cursor:
            row_data = []
            for col_data in rows:
                row_data.append(str(col_data)[0:14])
            print(row_format.format(*row_data))

    def compare_table_to_dataset(self, table_name, dataset_name):
        self.table_sample(table_name)
        self.dataset_sample(dataset_name)

    def build_schema_dataset(self, schema_name, use_comments=True, use_infer=False, use_sample_data=False, alternate_setname=None):
        self.m("Create API dataset for schema [" + schema_name + "]. Infer=" + str(use_infer) + " Sampledata=" + str(use_sample_data) + ".", tolog=True)
        # Prepare object for superset JSON registration
        superset_json = {}
        if not alternate_setname:
            superset_json["Setname"] = schema_name
        else:
            superset_json["Setname"] = alternate_setname
        superset_json["tables"] = []
        # Grab the tables in the correct order.
        working_cursor = self.connection.cursor()
        hierachy_stmt = self.load_resource("postgres_table_hierachy_schema.sql")
        working_cursor.execute(hierachy_stmt, (schema_name, schema_name,))
        for tab, dependson, level in working_cursor:
            # No matter if we have an existing dataset or not
            # We need to add to the superset
            table_def = {}
            table_def["table_name"] = tab
            table_def["table_level"] = level
            table_def["table_size"] = self.tableSize(tab)
            # Add dependents for size ratio calculations
            an_array = []
            depends_cursor = self.connection.cursor()
            depends_stmt = self.load_resource("postgres_foreign_key_depends.sql")
            depends_cursor.execute(depends_stmt, (tab,))
            for dep_tab in depends_cursor:
                an_object = {}
                an_object["local_col"] = dep_tab[2]
                an_object["remote_dataset"] = dep_tab[0]
                an_object["remote_col"] = dep_tab[1]
                an_array.append(an_object)    
            table_def["depends_on"] = an_array
            superset_json["tables"].append(table_def)
            if not self.dataset_exists(tab):
                self.logger.info("Building dataset for [%s] at level [%s].", tab, level)
                self.dataset_from_table(tab, use_comments, use_infer, use_sample_data)
            else:
                self.logger.info("Dataset for [%s] already exists", tab)
        # Now we have the whoel superset. Ship and create
        # TODO: Send to superset register service.
        self.logger.info("Superset data to store: %s", json.dumps(superset_json))
        superset_result = self.create_superset(json.dumps(superset_json))
        self.m("API dataset for schema [" + schema_name + "] successfully created.", tolog=True)


import csv
import logging
import json
import os
import datetime as dt
from statistics import mean
import decimal
from pathlib import Path
from binubuo import binubuo
from binubuo.BinubuoTemplate import BinubuoTemplate

class binubuoCSV(binubuo):
    def __init__(self, binubuokey):
        super().__init__(binubuokey)
        self.binubuokey = binubuokey
        self.max_sample_size = 100
        self.connected = False
        self.show_messages = True

    def setup_logging(self):
        # Create the logger
        self.logger = logging.getLogger(__name__)
        # Logger settings and levels
        self.logger.setLevel(logging.DEBUG)
        # Logger formatter and handler
        file_handler = logging.FileHandler(filename='binubuo.log', mode='a', encoding='utf8')
        formatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')
        file_handler.setFormatter(formatter)
        # Add the handler to the logger
        self.logger.addHandler(file_handler)

    def csv_line_count(self, fname):
        with open(fname) as f:
            return sum(1 for line in f)

    def can_float(self, x):
        try:
            float(x)
            return True
        except:
            return False

    def is_date_string(self, str):
        date_fmts = ('%Y-%m-%dT%H:%M:%S.%f', '%Y-%m-%d %H:%M:%S', '%d-%m-%Y %H:%M:%S','%Y-%m-%d', '%d-%m-%Y', '%d/%m/%Y', '%Y','%b %d, %Y','%b %d, %Y','%B %d, %Y','%B %d %Y','%m/%d/%Y','%m/%d/%y','%b %Y','%B%Y','%b %d,%Y')
        parsed = []
        if str is not None:
            for fmt in date_fmts:
                try:
                    t = dt.datetime.strptime(str, fmt)
                    parsed.append((str, fmt, t)) 
                    break
                except ValueError as err:
                    pass
        return parsed

    def get_file_csv_dialect(self, fname):
        try:
            self.logger.debug("Find dialect for %s", fname)
            with open(fname, newline='', encoding='utf-8') as csvfile:
                self.current_dialect = csv.Sniffer().sniff(csvfile.read(1024))
                csvfile.seek(0)
                self.logger.debug("Dialect found for %s", fname)
                self.logger.debug("CSV Dialect - Delimiter: '%s', Doublequote: '%s', Escapechar: '%s', Lineterminator: '%s', Quotechar: '%s'", self.current_dialect.delimiter, str(self.current_dialect.doublequote), self.current_dialect.escapechar, self.current_dialect.lineterminator, self.current_dialect.quotechar)
                # Here we need to see if we are actually quoting, and set the corresponding dialect attribute.
                # csv.QUOTE_ALL : Writer to quote all fields.
                # csv.QUOTE_MINIMAL : Writer to quote fields that contain special characters
                # csv.QUOTE_NONNUMERIC: Writer quotes all fields except numeric fields.
                # csv.QUOTE_NONE: Writer does not quote any fields and no special processing is required from the Reader object.
                row_length = None
                count_quotechars = None
                count = 0
                while True:
                    count += 1
                    line = csvfile.readline()
                    row_length = line.count(self.current_dialect.delimiter) + 1
                    row_quote_count = line.count(self.current_dialect.quotechar)/2
                    self.logger.debug('Row element lenght: %s, Quoted element count: %s for row [%s]', row_length, row_quote_count, line)
                    if count > 10:
                        break
                    if row_quote_count >= row_length:
                        count_quotechars = csv.QUOTE_ALL
                    elif row_quote_count > 0 and row_quote_count < row_length:
                        count_quotechars = csv.QUOTE_NONNUMERIC
                    else:
                        count_quotechars = csv.QUOTE_NONE
                # Now set the value for the dialect
                self.logger.debug("Quoting set as: %s", count_quotechars)
                self.current_dialect.quoting = count_quotechars
        except Exception as e:
            self.logger.error("Error %s happened during read.", e)

    def templateFromFile(self, file_name, known_header=None, output_type="json"):
        self.m("Create template from csv file [" + file_name + "].", tolog=True)
        table_template = BinubuoTemplate()
        # Open file and go through it.
        try:
            with open(file_name, newline='', encoding='utf-8') as csvfile:
                self.logger.info("File %s opened successfully", file_name)
                # Get the row count
                row_count = self.csv_line_count(file_name)
                self.logger.info("File %s has %s rows.", file_name, row_count)
                # Extract metadata
                self.current_dialect = csv.Sniffer().sniff(csvfile.read(1024))
                # Reset file pointer
                csvfile.seek(0)
                if known_header is not None:
                    self.current_has_header = known_header
                else:
                    self.current_has_header = csv.Sniffer().has_header(csvfile.read(1024))
                # Reset file pointer
                csvfile.seek(0)
                self.current_reader = csv.reader(csvfile, self.current_dialect)
                csv_sample_store = {}
                self.logger.debug("Start loading the sample store.")
                for count, row in enumerate(self.current_reader):
                    # Sample data store setup
                    if count == 0:
                        if self.current_has_header:
                        # Setup data store with header names
                            self.logger.debug("setting up data store with named headers")
                            for head_count, head_value in enumerate(row):
                                self.logger.debug("Creating store with header %s", head_value)
                                csv_sample_store[head_value] = []
                        else:
                            # Setup data stor with generated header names
                            self.logger.debug("setting up data store with generated headers")
                            for head_count, head_value in enumerate(row):
                                head_name = "c" + str(head_count)
                                self.logger.debug("Creating store with header %s", head_name)
                                csv_sample_store[head_name] = []
                    # Loop over rows. 
                    if count == int(self.binubuo_config.get('CSV', 'max_row_sample', fallback=100)):
                        break
                    else:
                        # We are in the sampling count, so split out the values to the columns
                        if self.current_has_header and count == 0:
                            pass
                        else:
                            for col_count, col_name in enumerate(csv_sample_store):
                                if not row[col_count]:
                                    csv_sample_store[col_name].append(None)
                                else:
                                    csv_sample_store[col_name].append(row[col_count])
            # Now we have the sample store. Enumerate each to get infer data.
            self.logger.debug("Enumerate sample store for infer JSON generation")
            for count, idx in enumerate(csv_sample_store):
                self.m("Sampling data on column [" + idx +"].", indent=True, tolog=False)
                self.logger.debug("Sample data for col %s: %s", idx, csv_sample_store[idx])
                infer_meta = {}
                self.logger.debug("Start creating infer JSON for column %s", idx)
                infer_meta["column_name"] = idx
                infer_meta["table_name"] = Path(file_name).stem
                # Nulls before checking for datatype
                self.logger.debug("before null check")
                infer_meta["column_nulls"] = sum(x is None for x in csv_sample_store[idx])
                if infer_meta["column_nulls"] > 0:
                    infer_meta["column_nullable"] = "Y"
                else:
                    infer_meta["column_nullable"] = "N"
                # Datatype check
                nums_chk = [x for x in csv_sample_store[idx] if self.can_float(x)]
                self.logger.debug("Numbers checked: %s", nums_chk)
                self.logger.debug("before datatype check")
                if len(nums_chk) == len(csv_sample_store[idx]) or len(nums_chk) == (len(csv_sample_store[idx]) - infer_meta["column_nulls"]):
                    infer_meta["column_datatype"] = "number"
                    infer_meta["column_number_decimals"] = round(mean([decimal.Decimal(str(x)).as_tuple().exponent for x in nums_chk])*-1)
                else:
                    # TODO: Here we need to check if it is a date, before we just set it to string
                    # Checking for date format is very intensive. So as soon as we have one fail (len(0)) returned from the checker, we give up.
                    is_date = None
                    date_format = None
                    for str_num, str_val in enumerate(csv_sample_store[idx]):
                        self.logger.debug("About to date check %s", str_val)
                        has_date_format = self.is_date_string(str_val)
                        if len(has_date_format) == 0:
                            is_date = False
                            break
                        else:
                            is_date = True
                            date_format = has_date_format[0][1]
                            self.logger.debug("Found format: %s", date_format)
                    if is_date:
                        infer_meta["column_datatype"] = "date"
                    else:
                        infer_meta["column_datatype"] = "string"
                self.logger.debug("before data metadata settings")
                infer_meta["table_level_rowcount"] = row_count
                infer_meta["table_level_avg_row_length"] = ""
                infer_meta["column_level_avg_col_length"] = round(sum( map(len, map(str, csv_sample_store[idx])) ) / len(csv_sample_store[idx]))
                infer_meta["column_data_length"] = "4000"
                infer_meta["column_number_precision"] = ""
                # Unique list
                u_set = set(csv_sample_store[idx])
                u_list = (list(u_set))
                # Bug with unique. For now ignore
                #if len(u_list) == len(csv_sample_store[idx]):
                #    infer_meta["column_is_unique"] = "yes"
                #    infer_meta["column_distinct_count"] = len(csv_sample_store[idx])
                #else:
                infer_meta["column_is_unique"] = "no"
                infer_meta["column_distinct_count"] = len(csv_sample_store[idx]) - len(u_list)
                self.logger.debug("before data low/high settings")
                if infer_meta["column_datatype"] == "number":
                    infer_meta["column_low_value"] = min([float(x) for x in nums_chk])
                    infer_meta["column_high_value"] = max([float(x) for x in nums_chk])
                elif infer_meta["column_datatype"] == "date":
                    self.logger.debug("Just before date min/max")
                    infer_meta["column_low_value"] = min([dt.datetime.strptime(x, date_format) for x in csv_sample_store[idx] if x is not None]).strftime('%Y-%m-%d %H:%M:%S')
                    infer_meta["column_high_value"] = max([dt.datetime.strptime(x, date_format) for x in csv_sample_store[idx] if x is not None]).strftime('%Y-%m-%d %H:%M:%S')
                elif infer_meta["column_datatype"] == "string":
                    infer_meta["column_low_value"] = min([x for x in csv_sample_store[idx] if x is not None])
                    infer_meta["column_high_value"] = max([x for x in csv_sample_store[idx] if x is not None])
                if infer_meta["column_datatype"] == "number":
                    infer_meta["sample_values"] = nums_chk
                else:
                    infer_meta["sample_values"] = csv_sample_store[idx]
                self.logger.debug("Infer: %s", json.dumps(infer_meta))
                table_template.init_column(column_name=idx)
                x_response = self.infer_generator(json.dumps(infer_meta))
                self.logger.info("Infer result: %s", x_response)
                table_template.replace_column_from_json(idx, json.dumps(x_response))
            self.m("Template for csv file [" + file_name +"] successfully created.", indent=True, tolog=False)
            if output_type.upper() == "JSON":
                table_template.validate_template()
                table_template.complete_template()
                return table_template.template_JSON
            else:
                return table_template
        except Exception as e:
            self.logger.error("Error %s happened during read.", e)

    def dataset_from_csv(self, file_name, known_header=None):
        dset_name = Path(file_name).stem.lower()
        dset_template = self.templateFromFile(file_name, known_header)
        self.create_dataset(dset_name, dset_template)

    def copy_csv(self, file_name, new_name
            , drop_target_if_exist=False, data_rows="source"
            , known_header=None):
        self.m("Create copy [" + new_name + "] of CSV  [" + file_name + "].", tolog=True)
        dset_name = Path(file_name).stem
        can_create = False
        self.logger.debug("Before checking existence")
        if not self.dataset_exists(dset_name):
            self.dataset_from_csv(file_name, known_header)
        # Always get fresh dialect.
        self.logger.debug("Before contained dialect check")
        self.get_file_csv_dialect(file_name)
        self.logger.debug("Received dialect and quoting is: %s", self.current_dialect.quoting)
        self.logger.debug("Check if %s already exists", new_name)
        if os.path.exists(new_name):
            if drop_target_if_exist:
                # remove file
                self.logger.debug("%s already exists, but allow dropped set. Removing file.", new_name)
                os.remove(new_name)
                can_create = True
        else:
            self.logger.debug("%s does not exists.", new_name)
            can_create = True
        if can_create:
            # Set the rows
            if data_rows == "source":
                self.drows(self.csv_line_count(file_name))
            else:
                self.drows(int(data_rows))
            # Get the data
            csv_data = self.dataset(dset_name)
            with open(new_name, 'w', encoding='UTF8', newline='') as fp:
                writer = csv.writer(fp, self.current_dialect)
                # TODO: Handle header
                self.logger.debug("About to write %s rows in batch mode to %s", len(csv_data), new_name)
                writer.writerows(csv_data)
            self.m("Copy " + new_name + " created successfully.", tolog=True)



import cx_Oracle
import sys
import json
import ast
import logging
from binubuo import binubuo
from binubuo.BinubuoTemplate import BinubuoTemplate

class binubuoOracle(binubuo):
    def __init__(self, binubuokey, dbuser, dbpwd, dbdsn, dbconfig=None, dbfullConnection=None):
        super().__init__(binubuokey)
        self.binubuokey = binubuokey
        self.dbuser = dbuser
        self.dbpwd = dbpwd
        self.dbdsn = dbdsn
        self.dbconfig = dbconfig
        self.dbfullConnection = dbfullConnection
        self.max_sample_size = 100
        self.connected = False
        self.show_messages = True

    def setup_logging(self):
        # Create the logger
        self.logger = logging.getLogger(__name__)
        # Logger settings and levels
        self.logger.setLevel(logging.DEBUG)
        # Logger formatter and handler
        file_handler = logging.FileHandler(filename='binubuo.log', mode='a')
        formatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')
        file_handler.setFormatter(formatter)
        # Add the handler to the logger
        self.logger.addHandler(file_handler)

    def connect(self):
        try:
            if self.dbfullConnection is None:
                if self.dbconfig is not None:   
                    cx_Oracle.init_oracle_client(config_dir=self.dbconfig)
                self.connection = cx_Oracle.connect(user=self.dbuser, password=self.dbpwd, dsn=self.dbdsn)
            else:
                self.connection = self.dbfullConnection
            self.connected = True
            # Set defaults
            working_cursor = self.connection.cursor()
            stmt = self.load_resource("oracle_nls_date_format.sql")
            self.logger.info(stmt)
            working_cursor.execute(stmt)
            stmt = self.load_resource("oracle_nls_timestamp_format.sql")
            self.logger.info(stmt)
            working_cursor.execute(stmt)
            stmt = self.load_resource("oracle_nls_timestamp_tz_format.sql")
            self.logger.info(stmt)
            working_cursor.execute(stmt)
        except:
            self.logger.error("Error: Failed to connect to database. Please make sure you have a connection before using other methods.")
            self.connection = None

    def connectSecondary(self, dbuser, dbpwd, dbdsn, dbconfig=None, dbfullConnection=None):
        try:
            if dbfullConnection is None:
                if dbconfig is not None:
                    try: 
                        cx_Oracle.init_oracle_client(config_dir=dbconfig)
                    except:
                        # If Client Library already loaded (with the primary connection) ignore the error here
                        pass
                self.secondary_connection = cx_Oracle.connect(user=dbuser, password=dbpwd, dsn=dbdsn)
            else:
                self.secondary_connection = dbfullConnection
            self.connected_secondary = True
            working_cursor = self.secondary_connection.cursor()
            stmt = self.load_resource("oracle_nls_date_format.sql")
            self.logger.info(stmt)
            working_cursor.execute(stmt)
            stmt = self.load_resource("oracle_nls_timestamp_format.sql")
            self.logger.info(stmt)
            working_cursor.execute(stmt)
            stmt = self.load_resource("oracle_nls_timestamp_tz_format.sql")
            self.logger.info(stmt)
            working_cursor.execute(stmt)
        except:
            self.logger.error("Error: Failed to connect to secondary/alternate database. Please check your connection details and try again.")
            self.secondary_connection = None

    def calculateSampleSize(self, table_name):
        sub_sample_cursor = self.connection.cursor()
        sub_sample_size_stmt = self.load_resource("oracle_table_rows.sql")
        sub_sample_cursor.execute(sub_sample_size_stmt, tabname = table_name.upper())
        sub_sample_size_cal = sub_sample_cursor.fetchone()
        # Calculate the real sample size in percent
        if (int(sub_sample_size_cal[0]) > 0 and int(sub_sample_size_cal[0]) < self.max_sample_size):
            sub_sample_size_cal_cust = 99.99
        elif (int(sub_sample_size_cal[0]) > self.max_sample_size):
            # Calculate rough percentage of max sample size
            sub_sample_size_cal_cust = round((self.max_sample_size/int(sub_sample_size_cal[0])*100), 8)
        else:
            # In case of zero rows, we read all of nothing :)
            sub_sample_size_cal_cust = 99.99
        self.logger.info("Sample size of %s calculated for %s rows.", str(sub_sample_size_cal_cust), str(sub_sample_size_cal[0]))
        return sub_sample_size_cal_cust

    def tableSize(self, table_name):
        table_size_cursor = self.connection.cursor()
        table_size_size_stmt = self.load_resource("oracle_table_rows.sql")
        table_size_cursor.execute(table_size_size_stmt, tabname = table_name.upper())
        table_size_size_cal = table_size_cursor.fetchone()
        table_size_size_cal = table_size_size_cal[0]
        self.logger.info("Table size of %s is %s rows.", table_name, str(table_size_size_cal))
        return table_size_size_cal

    def templateFromTable(self, table_name, use_comments=True, use_infer=False, use_sample_data=False, output_type="json"):
        self.m("Create template from table [" + table_name + "]. Infer=" + str(use_infer) + " Sampledata=" + str(use_sample_data) + ".", tolog=True)
        table_template = BinubuoTemplate()
        # Let us validate the table name before we continue.
        table_exist = self.table_exist(table_name)
        # There are a number of different ways to branch out and do the template.
        # 1. Just use generic defaults for datatype.
        # 2. Use comments if available, and fall back to defaults if a column does not have a comment.
        # 3. Call the webservice with _only_ metadata to infer the generator
        # 4. Call the webservice with metadata _and_ sample data to infer the generator.
        if table_exist:
            self.templateTable = table_name.upper()
            if not use_infer:
                self.m("Using simple method to create columns.", indent=True, tolog=False)
                # No webservice call to generate template.
                working_cursor = self.connection.cursor()
                generic_column_meta_sql = self.load_resource("oracle_column_meta_simple.sql")
                working_cursor.execute(generic_column_meta_sql, tabname = table_name.upper())
                for cname, cdtype, cdgenerator, ccomment, crows, cnumnulls, cnumdist, cnullable, cavglen in working_cursor:
                    table_template.init_column(column_name=cname, column_type="generated", column_datatype=cdtype)
                    if ccomment is not None and use_comments:
                        table_template.set_column_attribute(cname, "generator", ccomment)
                    else:
                        table_template.set_column_attribute(cname, "generator", cdgenerator)
                # Once we are done, we can validate the template and return it.
                self.m("Template for table [" + table_name +"] successfully created.", indent=True, tolog=False)
                if output_type.upper() == "JSON":
                    table_template.validate_template()
                    table_template.complete_template()
                    return table_template.template_JSON
                else:
                    return table_template
            else:
                # Placeholder to run the infer code.
                self.m("Using infer method to create columns.", indent=True, tolog=False)
                if use_sample_data:
                    sub_sample_size_cal_cust = self.calculateSampleSize(table_name)
                working_cursor = self.connection.cursor()
                if use_sample_data:
                    columns_meta_stmt = self.load_resource("oracle_column_meta_plus_sample.sql")
                else:
                    columns_meta_stmt = self.load_resource("oracle_column_meta_plus_standard.sql")
                self.logger.debug("Statement to get column data: %s", columns_meta_stmt)
                working_cursor.execute(columns_meta_stmt, tabname = table_name.upper())
                for col_name, col_infer_meta in working_cursor:
                    self.logger.info("Column to infer: %s", col_name)
                    # If we are allowed sample data. Now is the time to fetch individual samples
                    if use_sample_data:
                        self.m("Sampling data on column [" + col_name +"].", indent=True, tolog=False)
                        sub_sample_cursor = self.connection.cursor()
                        # Merge the sample data.
                        # First parse the main column metadata
                        main_meta = json.loads(col_infer_meta)
                        sub_sample_stmt = self.load_resource("oracle_column_sample_data.sql").format(r_col_name=col_name, r_table_name=table_name, r_table_sample_size=sub_sample_size_cal_cust)
                        self.logger.debug("Statement to get column (%s) sample data: %s", col_name, sub_sample_stmt)
                        sub_sample_cursor.execute(sub_sample_stmt)
                        sample_data = sub_sample_cursor.fetchone()
                        # First we get the sample data as a clob string
                        sample_data_clob_str = sample_data[0].read()
                        # Convert to a list so we can add for real.
                        sample_data_clob_str = ast.literal_eval(sample_data_clob_str)
                        # Now add the sample data as a real array.
                        main_meta["sample_values"] = sample_data_clob_str
                        if main_meta["column_is_reference"] == 1:
                            self.m("Column [" + col_name +"] is foreign key. Create key distribution histogram.", indent=True, tolog=False)
                            rel_sample_size_cal_cust = self.calculateSampleSize(main_meta["reference_table"])
                            rel_sample_cursor = self.connection.cursor()
                            rel_sample_stmt = self.load_resource("oracle_foreign_key_histogram.sql").format(col_name = main_meta["column_name"], tab_name = table_name, ref_col_name = main_meta["reference_column"], ref_tab_name = main_meta["reference_table"], ref_sample = rel_sample_size_cal_cust)
                            self.logger.info("Column is a foreign key. Build relation histogram")
                            self.logger.debug("Relation histogram query: %s", rel_sample_stmt)
                            rel_sample_cursor.execute(rel_sample_stmt)
                            rel_sample_data = rel_sample_cursor.fetchone()
                            self.logger.debug("Foreign key histogram: ")
                            rel_sample_data_clob_str = rel_sample_data[0].read()
                            self.logger.debug(rel_sample_data_clob_str)
                            rel_sample_data_clob_str = ast.literal_eval(rel_sample_data_clob_str)
                            main_meta["reference_histogram"] = rel_sample_data_clob_str
                    else:
                        main_meta = json.loads(col_infer_meta)
                    # We now have every column and its metadata. Send to API endpoint to do intelligent infer.
                    # We get back the entire json of inferred column
                    # Init column
                    table_template.init_column(column_name=col_name)
                    # Try calling infer endpoint
                    self.logger.info("Infer: %s", json.dumps(main_meta))
                    x_response = self.infer_generator(json.dumps(main_meta))
                    self.logger.info(x_response)
                    # The response is the column json as a string format. Call replace_column_from_json
                    table_template.replace_column_from_json(col_name, json.dumps(x_response))
                self.m("Template for table [" + table_name +"] successfully created.", indent=True, tolog=False)
                # Once we are done, we can validate the template and return it.
                if output_type.upper() == "JSON":
                    table_template.validate_template()
                    table_template.complete_template()
                    return table_template.template_JSON
                else:
                    return table_template

    def columnGeneratorComments(self, table_name, comments, override=False):
        self.m("Create comments on table [" + table_name + "]. Comments=" + str(comments) + ".", tolog=True)
        if self.connected:
            comment_count_sql = self.load_resource("oracle_count_comments.sql")
            working_cursor = self.connection.cursor()
            working_cursor.execute(comment_count_sql, tabname = table_name.upper())
            cursor_result = working_cursor.fetchone()
            if (cursor_result[0] > 0 or not override) or (cursor_result[0] == 0):
                # Let us go ahead and set comment generators as requested.
                if (comments.find('=') > 0):
                    # Named notation
                    for cn in comments.split(","):
                        colname = cn.split("=")[0].strip()
                        colcomment = cn.split("=")[1].strip()
                        col_add_stmt = "comment on column " + table_name + "." + colname + " is '" + colcomment + "'"
                        working_cursor.execute(col_add_stmt)
                else:
                    # Just split and follow column order. Ignore any column outside of split length
                    ordered_col_cursor = self.connection.cursor()
                    ordered_col_sql = self.load_resource("oracle_comments_ordered.sql")
                    ordered_col_cursor.execute(ordered_col_sql, tabname = table_name.upper())
                    for cid, cname in ordered_col_cursor:
                        try:
                            colcomment = comments.split(",")[cid - 1]
                            col_add_stmt = "comment on column " + table_name + "." + cname + " is '" + colcomment + "'"
                            working_cursor.execute(col_add_stmt)
                        except:
                            # Aint no comment for this column. Just ignore
                            pass
                self.m("Comments for table [" + table_name +"] successfully created.", indent=True, tolog=False)

    def quick_fetch_table(self
            , table_name, use_comments=True, use_infer=False
            , use_sample_data=False, use_tuple_return=False
            , output_csv=False, output_type="screen", output_name="same"
            , rows="source"):
        self.template = BinubuoTemplate()
        table_exist = self.table_exist(table_name)
        # Table exists
        if table_exist:
            quick_fetch_columns_array = []
            if not use_infer:
                # No webservice call to generate template.
                working_cursor = self.connection.cursor()
                generic_column_meta_sql = self.load_resource("oracle_column_meta_simple.sql")
                working_cursor.execute(generic_column_meta_sql, tabname = table_name.upper())
                for cname, cdtype, cdgenerator, ccomment, crows, cnumnulls, cnumdist, cnullable, cavglen in working_cursor:
                    self.template.init_column(column_name=cname, column_type="generated", column_datatype=cdtype)
                    if ccomment is not None and use_comments:
                        quick_fetch_columns_array.append(ccomment)
                    else:
                        quick_fetch_columns_array.append(cdgenerator)
                # Once we are done, we can call the quick fetch.
                quick_fetch_columns = ",".join(quick_fetch_columns_array)
                # Before fetching, set rows if needed.
                if rows == "source":
                    self.drows(self.tableSize(table_name))
                else:
                    try:
                        self.drows(int(rows))
                    except:
                        self.drows()
                if use_tuple_return or output_type.lower() == "table":
                    tuple_value = "tuple"
                else:
                    tuple_value = "list"
                if output_csv:
                    self.csv(1)
                if output_name == "same":
                    output_name_real = table_name
                else:
                    output_name_real = output_name
                if output_type.lower() == "screen" or output_type.lower() == "table":
                    resp_cols = self.quick_fetch(quick_fetch_columns, tuple_value)
                    self.logger.info("Length is: %s", str(len(resp_cols)))
                    self.logger.info(resp_cols)
                    if output_type.lower() == "screen":
                        return resp_cols
                    else:
                        # We need to insert.
                        quick_insert_stmt = self.build_insert_statement(table_name, output_name_real)
                        quick_insert_cursor = self.connection.cursor()
                        bulk_limit = int(self.binubuo_config.get('DATABASE', 'oracle_bulk_size', fallback='400'))
                        if len(resp_cols) < bulk_limit:
                            # One execute
                            self.logger.info("Bulk limit [%s] is higher than actual row count [%s]. No need to split executemany.", str(bulk_limit), str(len(resp_cols)))
                            quick_insert_cursor.executemany(quick_insert_stmt, resp_cols)
                            self.logger.info("Rows inserted: %s", str(len(resp_cols)))
                        else:
                            # Split up to smaller parts
                            self.logger.info("Bulk limit [%s] is lower than actual row count [%s]. Split to multiple executemany.", str(bulk_limit), str(len(resp_cols)))
                            splitted_arrays = [resp_cols[i:i+bulk_limit] for i in range(0, len(resp_cols), bulk_limit)]
                            for bulks in splitted_arrays:
                                quick_insert_cursor.executemany(quick_insert_stmt, bulks)
                                self.logger.info("Rows inserted: %s", str(len(bulks)))
                        self.connection.commit()
                elif output_type.lower() == "file":
                    self.quick_fetch_to_file(quick_fetch_columns, output_name_real)
                # Reset required vals
                self.csv()
                self.drows()

    def dataset_from_table(self, table_name, use_comments=True, use_infer=False, use_sample_data=False):
        dset_template = self.templateFromTable(table_name, use_comments, use_infer, use_sample_data)
        # Create or replace the dataset
        self.create_dataset(table_name, dset_template)

    def build_insert_statement(self, source_table, target_table=None):
        insert_stmt_cursor = self.connection.cursor()
        build_insert_stmt = self.load_resource("oracle_build_insert.sql")
        insert_stmt_cursor.execute(build_insert_stmt, tabname=source_table.upper())
        cursor_result = insert_stmt_cursor.fetchone()
        if target_table is not None:
            insert_stmt = "insert into " + target_table.lower() + " " + cursor_result[0]
        else:
            insert_stmt = "insert into " + source_table.lower() + " " + cursor_result[0]
        self.logger.debug("Insert stmt: %s", insert_stmt)
        return insert_stmt

    def table_exist(self, table_name, use_secondary=False):
        if use_secondary and self.connected_secondary:
            use_conn = self.secondary_connection
        else:
            use_conn = self.connection
        assert_table_sql = self.load_resource("oracle_assert_table.sql")
        # SOURCE work
        assert_table_cursor = use_conn.cursor()
        table_exist = False
        try:
            assert_table_cursor.execute(assert_table_sql, tabname = table_name.upper())
            table_exist = True
            self.logger.info("Table %s asserted and it exists", table_name)
        except:
            table_exist = False
            self.logger.warning("Table %s asserted and it does not exists", table_name)
        return table_exist

    def ddl_from_table(self, table_name):
        ddl_cursor = self.connection.cursor()
        ddl_stmt = self.load_resource("oracle_prepare_dbms_metadata.sql")
        # Set the transformation rules for DDL extraction.
        ddl_cursor.execute(ddl_stmt)
        # Extract DDL
        ddl_stmt = self.load_resource("oracle_table_ddl.sql")
        ddl_cursor.execute(ddl_stmt, tabname = table_name.upper())
        ddl_out_str = ddl_cursor.fetchone()
        ddl_out_str = ddl_out_str[0].read()
        return ddl_out_str

    def copy_table(self, source_table, target_table=None, copy_method="quickfetch"
            , drop_target_if_exist=False, alternate_dataset_name=False
            , use_comments=True, use_infer=False, use_sample_data=False
            , data_rows="source", target_db=None):
        working_cursor = self.connection.cursor()
        if target_db is None:
            target_work_cursor = self.connection.cursor()
        else:
            target_work_cursor = self.secondary_connection.cursor()
        source_table_exist = self.table_exist(source_table)
        # TARGET work
        # If target_table is none, we will create a copy with same name plus '_copy' appended to it.
        if target_table is None:
            loc_target_table = source_table + "_copy"
        else:
            loc_target_table = target_table
        self.m("Create copy [" + loc_target_table + "] of table [" + source_table + "]. Infer=" + str(use_infer) + " Sampledata=" + str(use_sample_data) + ".", tolog=True)
        target_table_exist = self.table_exist(loc_target_table) if target_db is None else self.table_exist(loc_target_table, True)
        if target_table_exist:
            if drop_target_if_exist:
                remove_target_stmt = "drop table " + loc_target_table.lower() + " purge"
                self.logger.debug("Dropping target %s with stmt: %s", loc_target_table, remove_target_stmt)
                target_work_cursor.execute(remove_target_stmt)
                target_table_exist = False
            else:
                self.logger.error("Target table exist cannot create.")
        if source_table_exist and not target_table_exist:
            # Source is there and target is not there. Good to go.
            # First we create an empty copy table
            # If local database we can use create as select ...
            # If secondary database, we do extract DDL
            if target_db is None:
                table_ddl = "create table " + loc_target_table.lower() + " as select * from " + source_table.lower() + " where 1=2"
            else:
                # We need to extract from primary the stmt to build in secondary
                table_ddl = self.ddl_from_table(source_table.upper())
                # Now we just need to replace tablename, since the DDL will be source table name
                table_ddl = table_ddl.replace('"' + source_table.upper() + '"', '"' + loc_target_table.upper() + '"')
            self.logger.debug("Create table stmt: %s", table_ddl)
            target_work_cursor.execute(table_ddl)
            # Table created. Build insert stmt
            insert_stmt = self.build_insert_statement(source_table, loc_target_table)
            # Set the rows
            if data_rows == "source":
                # Get the rowcount in the source table.
                self.drows(self.tableSize(source_table))
            else:
                try:
                    self.drows(int(data_rows))
                except:
                    self.drows(int(10))
            # Get new rows.
            if copy_method == "quickfetch":
                rows_bind = self.quick_fetch_table(table_name=source_table, use_tuple_return=True)
                if rows_bind is not None:
                    self.logger.info("Rows returned for insert: %s", str(len(rows_bind)))
                    target_work_cursor.executemany(insert_stmt, rows_bind)
            elif copy_method == "dataset":
                dset_name = source_table if not alternate_dataset_name else alternate_dataset_name
                if not self.dataset_exists(dset_name):
                    self.dataset_from_table(dset_name, use_comments, use_infer, use_sample_data)
                # Call the dataset for the rows.
                rows_bind = self.dataset(dataset_name=dset_name, response_type="tuple")
            # Insert the rows
            if rows_bind is not None:
                self.logger.info("Rows returned for insert: %s", str(len(rows_bind)))
                bulk_limit = int(self.binubuo_config.get('DATABASE', 'oracle_bulk_size', fallback='400'))
                if len(rows_bind) < bulk_limit:
                    # One execute
                    self.logger.info("Bulk limit [%s] is higher than actual row count [%s]. No need to split executemany.", str(bulk_limit), str(len(rows_bind)))
                    target_work_cursor.executemany(insert_stmt, rows_bind)
                    self.logger.info("Rows inserted: %s", str(len(rows_bind)))
                else:
                    # Split up to smaller parts
                    self.logger.info("Bulk limit [%s] is lower than actual row count [%s]. Split to multiple executemany.", str(bulk_limit), str(len(rows_bind)))
                    splitted_arrays = [rows_bind[i:i+bulk_limit] for i in range(0, len(rows_bind), bulk_limit)]
                    for bulks in splitted_arrays:
                        target_work_cursor.executemany(insert_stmt, bulks)
                        self.logger.info("Rows inserted: %s", str(len(bulks)))
                if target_db is None:
                    self.connection.commit()
                else:
                    self.secondary_connection.commit()
            self.m("Synthetic table [" + loc_target_table +"] successfully created.", indent=True, tolog=False)

    def insert_from_dataset(self, dataset_name, table_name
        , data_rows="source", category="custom", **webargs):
        table_exist = self.table_exist(table_name)
        do_insert = False
        if table_exist:
            if data_rows == "source":
                # Get the rowcount in the source table.
                self.drows(self.tableSize(table_name))
            else:
                try:
                    self.drows(int(data_rows))
                except:
                    self.drows(int(10))
            # Create cursor and insert statement
            insert_cursor = self.connection.cursor()
            insert_stmt = self.build_insert_statement(table_name)
            if category == "custom":
                if self.dataset_exists(dataset_name):
                    # Custom dataset does not exists
                    rows_bind = self.dataset(dataset_name=dataset_name, response_type="tuple", **webargs)
                    do_insert = True
                elif self.standard_dataset_exists(dataset_name) is not None and self.binubuo_config['DATABASE']['use_standard_datasets_as_source'] == 'yes':
                    if self.binubuo_config.get('DATABASE', 'use_standard_category_if_duplicate', fallback=None) is None:
                        std_dataset_cat = self.standard_dataset_exists(dataset_name)
                    else:
                        std_dataset_cat = self.binubuo_config.get('DATABASE', 'use_standard_category_if_duplicate')
                    rows_bind = self.dataset(dataset_name=dataset_name, dataset_type="standard", dataset_category=std_dataset_cat, response_type="tuple", **webargs)
                    do_insert = True
                else:
                    self.logger.error("Dataset does not exist. Unable to insert.")
            else:
                if self.standard_dataset_exists(dataset_name, category).upper() == category.upper():
                    # We can do standard fetch
                    rows_bind = self.dataset(dataset_name=dataset_name, dataset_type="standard", dataset_category=category, response_type="tuple", **webargs)
                    do_insert = True
                else:
                    self.logger.error("Dataset does not exist. Unable to insert.")
            if do_insert and rows_bind is not None:
                self.logger.info("Rows returned for insert: %s", str(len(rows_bind)))
                bulk_limit = int(self.binubuo_config.get('DATABASE', 'oracle_bulk_size', fallback='400'))
                if len(rows_bind) < bulk_limit:
                    # One execute
                    self.logger.info("Bulk limit [%s] is higher than actual row count [%s]. No need to split executemany.", str(bulk_limit), str(len(rows_bind)))
                    insert_cursor.executemany(insert_stmt, rows_bind)
                    self.logger.info("Rows inserted: %s", str(len(rows_bind)))
                else:
                    # Split up to smaller parts
                    self.logger.info("Bulk limit [%s] is lower than actual row count [%s]. Split to multiple executemany.", str(bulk_limit), str(len(rows_bind)))
                    splitted_arrays = [rows_bind[i:i+bulk_limit] for i in range(0, len(rows_bind), bulk_limit)]
                    for bulks in splitted_arrays:
                        insert_cursor.executemany(insert_stmt, bulks)
                    self.logger.info("Rows inserted: %s", str(len(bulks)))
                self.connection.commit()
        else:
            self.logger.warning("Table does not exist. Unable to insert.")

    def table_sample(self, table_name):
        # Print title
        print("**************************************************************************")
        print("** Table data sample: " + table_name.upper())
        print("**************************************************************************")
        # Get header data
        headers = []
        headers_trimmed = []
        headers_deco = []
        header_stmt = self.load_resource("oracle_columns_ordered.sql")
        working_cursor = self.connection.cursor()
        working_cursor.execute(header_stmt, tabname = table_name.upper())
        for cname in working_cursor:
            headers.append(cname[0])
            headers_trimmed.append(cname[0][0:14])
        row_format = "{:>15}" * (len(headers))
        for x in range (len(headers)):
            headers_deco.append(" ==============")
        print(row_format.format(*headers_trimmed))
        print(row_format.format(*headers_deco))
        sample_stmt = self.load_resource("oracle_sample_select.sql").format(r_rows = ','.join(headers), r_table = table_name, r_limit = '3')
        working_cursor.execute(sample_stmt)
        for rows in working_cursor:
            row_data = []
            for col_data in rows:
                row_data.append(str(col_data)[0:14])
            print(row_format.format(*row_data))

    def compare_table_to_dataset(self, table_name, dataset_name):
        self.table_sample(table_name)
        self.dataset_sample(dataset_name)

    def build_schema_dataset(self, schema_name=None, use_comments=True, use_infer=False, use_sample_data=False, alternate_setname=None):
        # Prepare object for superset JSON registration
        working_cursor = self.connection.cursor()
        if schema_name is None:
            user_stmt = self.load_resource("oracle_current_user.sql")
            working_cursor.execute(user_stmt)
            for user in working_cursor:
                schema_name_set = user[0]
        superset_json = {}
        if not alternate_setname:
            superset_json["Setname"] = schema_name_set
        else:
            superset_json["Setname"] = alternate_setname
        self.m("Create API dataset for schema [" + schema_name_set + "]. Infer=" + str(use_infer) + " Sampledata=" + str(use_sample_data) + ".", tolog=True)
        superset_json["tables"] = []
        # Grab the tables in the correct order.
        hierachy_stmt = self.load_resource("oracle_table_hierachy_schema.sql")
        working_cursor.execute(hierachy_stmt)
        for tab, dependson, level in working_cursor:
            # No matter if we have an existing dataset or not
            # We need to add to the superset
            table_def = {}
            table_def["table_name"] = tab
            table_def["table_level"] = level
            table_def["table_size"] = self.tableSize(tab)
            # Add dependents for size ratio calculations
            an_array = []
            depends_cursor = self.connection.cursor()
            depends_stmt = self.load_resource("oracle_foreign_key_depends.sql")
            depends_cursor.execute(depends_stmt, tabname = tab.upper())
            for dep_tab in depends_cursor:
                an_object = {}
                an_object["local_col"] = dep_tab[2]
                an_object["remote_dataset"] = dep_tab[0]
                an_object["remote_col"] = dep_tab[1]
                an_array.append(an_object)    
            table_def["depends_on"] = an_array
            superset_json["tables"].append(table_def)
            if not self.dataset_exists(tab):
                self.logger.info("Building dataset for [%s] at level [%s].", tab, level)
                self.dataset_from_table(tab, use_comments, use_infer, use_sample_data)
            else:
                self.logger.info("Dataset for [%s] already exists", tab)
        # Now we have the whoel superset. Ship and create
        # TODO: Send to superset register service.
        self.logger.info("Superset data to store: %s", json.dumps(superset_json))
        superset_result = self.create_superset(json.dumps(superset_json))
        self.m("API dataset for schema [" + schema_name_set + "] successfully created.", tolog=True)

    def table_from_dataset(self, dataset_name, table_name=None, rows=0, category="custom", **webargs):
        self.m("Create table from dataset [" + dataset_name + "]. Number of rows=" + str(rows) + ".", tolog=True)
        self.logger.debug("Table from dataset using category: %s", category)
        # First we make sure that we have an up-to-date cache of datasets and metadata
        self.list_datasets(1)
        # Set the do_build
        do_build = False
        # Set the use standard
        use_standard = False
        # Now check if the dataset exists
        if category == "custom":
            if self.dataset_exists(dataset_name):
                do_build = True
            elif self.standard_dataset_exists(dataset_name) is not None and self.binubuo_config['DATABASE']['use_standard_datasets_as_source'] == 'yes':
                do_build = True
                use_standard = True
            else:
                self.m("Error: Dataset [" + dataset_name + "] does not exist.", tolog=True)
        else:
            if self.standard_dataset_exists(dataset_name, category).upper() == category.upper():
                do_build = True
                use_standard = True
            else:
                self.m("Error: Dataset [" + dataset_name + "] does not exist.", tolog=True)
        if do_build:
            if table_name is None:
                table_name_set = dataset_name
            else:
                table_name_set = table_name
            # Now check if the table exists
            if self.table_exist(table_name_set):
                self.logger.warning("Table [%s] does already exist. Either change name of requested table or drop existing.", table_name_set)
                # Here we should do "if allow drop, then drop"
            else:
                self.logger.info("About to create table [%s] from dataset definition", table_name_set)
                for idx, val in enumerate(self.dict_cache['datasets']):
                    for key, value in val.items():
                        if use_standard:
                            if key == "DATASET_WEBSERVICE_NAME" and value.lower() == dataset_name.lower():
                                if category == "custom":
                                    if self.dict_cache['datasets'][idx]["DATASET_TYPE_NAME"] == "Standard Dataset":
                                        dataset_cols = self.dict_cache['datasets'][idx]["DATASET_COLS"]
                                else:
                                    if self.dict_cache['datasets'][idx]["DATASET_TYPE_NAME"] == "Standard Dataset" and self.dict_cache['datasets'][idx]["DATASET_CATEGORY_NAME"].upper() == category.upper():
                                        dataset_cols = self.dict_cache['datasets'][idx]["DATASET_COLS"]
                        else:
                            if key == "DATASET_WEBSERVICE_NAME" and value.lower() == dataset_name.lower():
                                dataset_cols = self.dict_cache['datasets'][idx]["DATASET_COLS"]
                self.logger.info("Columns found for create table: %s", str(len(dataset_cols)))
                create_table_stmt = "create table " + table_name_set + " ("
                # Oracle datatype translation dictionary
                translation_dict = {'string': 'varchar2(4000)'
                    , 'time': 'timestamp'
                    , 'text': 'clob'}
                # loop over all columns and create the ddl for the table.
                cols_translated = []
                for idx, val in enumerate(dataset_cols):
                    trsl_datatype = dataset_cols[idx]["COL_DATATYPE"]
                    self.logger.debug("Add column [%s] of type [%s]", dataset_cols[idx]["COL_NAME"], trsl_datatype)
                    for word, replacement in translation_dict.items():
                        trsl_datatype = trsl_datatype.replace(word, replacement)
                    cols_translated.append(dataset_cols[idx]["COL_NAME"].lower() + " " + trsl_datatype)
                # Now all columns are done. Add to the statement
                all_cols = ','.join(cols_translated)
                create_table_stmt = create_table_stmt + all_cols + ")"
                self.logger.debug("Creating table: %s", create_table_stmt)
                working_cursor = self.connection.cursor()
                working_cursor.execute(create_table_stmt)
                self.m("Table " + table_name_set + " from dataset [" + dataset_name + "] created successfully.", tolog=True)
                try:
                    if rows > 0:
                        self.insert_from_dataset(dataset_name, table_name_set, rows, category, **webargs)
                except:
                    self.logger.debug("Either zero rows supplied or rows parameter not a number.")

    def create_schema_dataset(self, schema_name, scale=1, create_tables=True
        , create_copy=False, create_copy_suffix="_copy", data_tag="same_data"):
        self.m("Create superset from schema [" + schema_name + "]. At scale=" + str(scale) + ".", tolog=True)
        # If current tag is not null keep it, else set and reset back to none after.
        reset_tag = False
        if self.tag_set is None:
            self.tag(data_tag)
            reset_tag = True
        table_copies = []
        # First get the superset metadata
        superset = self.get_superset(schema_name, scale)
        # Get a session id and set it.
        session_id = self.create_session()
        session_id = session_id["bsession"]
        self.session_id(session_id)
        # Find out what type we are building
        superset_category = superset['settype'].lower()
        self.logger.debug("Category for superset: %s", superset_category)
        # Then traverse the datasets in order
        for idx, val in enumerate(superset['datasets']):
            dataset_name = superset['datasets'][idx]["dataset_name"]
            dataset_rows = superset['datasets'][idx]["scaled_size"]
            self.logger.info("Building dataset %s", dataset_name)
            # Build kwargs array for dataset
            dataset_kwargs = {}
            #if "dataset_inputs" in superset['datasets'][idx]:
            #    for idx2, val2 in enumerate(superset['datasets'][idx]["dataset_inputs"]):
            #        dataset_kwargs[superset['datasets'][idx]["dataset_inputs"][idx2]["input_id_name"]] = superset['datasets'][idx]["dataset_inputs"][idx2]["input_id_value"]
            if create_tables:
                if self.table_exist(dataset_name):
                    table_copies.append(dataset_name)
                    self.logger.debug("Adding %s to copy list for constraint replacement", dataset_name)
                    copy_name = dataset_name + create_copy_suffix
                    if create_copy and not self.table_exist(copy_name):
                        self.table_from_dataset(dataset_name=dataset_name, table_name=copy_name, rows=dataset_rows, category=superset_category)
                    else:
                        self.logger.warning("Failed to create copy table [%s] because copy table already exists", copy_name)
                else:
                    self.table_from_dataset(dataset_name=dataset_name, rows=dataset_rows, category=superset_category)
        # Now all datasets are tables. Let us add unqiue keys
        if create_tables:
            working_cursor = self.connection.cursor()
            # Add unqiue keys
            self.m("Adding primary keys.", True)
            for idx, val in enumerate(superset['unique keys']):
                exec_stmt = val.lower()
                self.logger.debug("Original constraint stmt: %s", exec_stmt)
                for x_val in table_copies:
                    fstr = "table " + x_val.lower() + " add"
                    self.logger.debug("Check if %s exists in %s", fstr, exec_stmt)
                    if exec_stmt.find(fstr) > 0:
                        # Replace val
                        exec_stmt = exec_stmt.replace(fstr, "table " + x_val.lower() + create_copy_suffix + " add")
                        exec_stmt = exec_stmt.replace(" unique", create_copy_suffix + " unique")
                self.logger.debug("Adding unique key: %s", exec_stmt)
                working_cursor.execute(exec_stmt)
            # Add foreign keys
            self.m("Adding foreign keys.", True)
            for idx, val in enumerate(superset['foreign keys']):
                exec_stmt = val.lower()
                for x_val in table_copies:
                    fstr1 = "table " + x_val.lower() + " add"
                    if exec_stmt.find(fstr1) > 0:
                        # Replace val
                        exec_stmt = exec_stmt.replace(fstr1, "table " + x_val.lower() + create_copy_suffix + " add")
                        exec_stmt = exec_stmt.replace(" foreign", create_copy_suffix + " foreign")
                    fstr2 = "references " + x_val.lower()
                    if exec_stmt.find(fstr2) > 0:
                        # Replace val
                        exec_stmt = exec_stmt.replace(fstr2, "references " + x_val.lower() + create_copy_suffix)
                self.logger.debug("Adding foreign key: %s", exec_stmt)
                try:
                    working_cursor.execute(exec_stmt)
                except:
                    self.logger.debug("Failed to add foreign key - %s", exec_stmt)
        if reset_tag:
            self.tag()
        session_id = self.remove_session()
        self.m("Superset from schema [" + schema_name + "] created successfully.", tolog=True)